#!/usr/bin/python
# p2p_service.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (p2p_service.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: p2p_service.

This serves requests from peers:

    * Data          - save packet to a file      (a commands.Data() packet)
    * Retrieve      - read packet from a file    (a commands.Data() packet)
    * ListFiles     - list files we have for customer
    * Delete        - delete a file
    * Identity      - contact or id server sending us a current identity
    * Ack           - response from remote peer after my request
    * Message       - a message from remote peer
    * Correspondent - request to be my correspondent
    TODO: describe other packets here

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.contacts import contactsdb

from bitdust.main import settings

from bitdust.p2p import commands

from bitdust.lib import packetid
from bitdust.lib import nameurl
from bitdust.lib import serialization

from bitdust.crypt import signed
from bitdust.crypt import my_keys

from bitdust.transport import gateway
from bitdust.transport import callback

from bitdust.userid import my_id

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.init')
    callback.insert_outbox_filter_callback(0, outbox)


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.shutdown')
    callback.remove_outbox_filter_callback(outbox)


#------------------------------------------------------------------------------


def inbox(newpacket, info, status, error_message):
    #     if newpacket.CreatorID != my_id.getIDURL() and newpacket.RemoteID != my_id.getIDURL():
    #         # packet is NOT for us, skip
    #         return False

    if newpacket.Command == commands.Identity():
        # a response from remote node, typically handled in other places
        Identity(newpacket, info)

    elif newpacket.Command == commands.Ack():
        # a response from remote node, typically handled in other places
        Ack(newpacket, info)

    elif newpacket.Command == commands.Fail():
        # some operation was failed on other side
        Fail(newpacket)

    elif newpacket.Command == commands.Retrieve():
        # retrieve some packet customer stored with us
        # handled by service_supplier()
        Retrieve(newpacket)

    elif newpacket.Command == commands.RequestService():
        # other node send us a request to get some service
        # handled by service_p2p_hookups()
        RequestService(newpacket, info)

    elif newpacket.Command == commands.CancelService():
        # other node wants to stop the service we gave him
        # handled by service_p2p_hookups()
        CancelService(newpacket, info)

    elif newpacket.Command == commands.Data():
        # new packet to store for customer, or data coming back from supplier
        # handled by service_backups() and service_supplier()
        Data(newpacket)

    elif newpacket.Command == commands.ListFiles():
        # customer wants list of their files
        # handled by service_supplier()
        ListFiles(newpacket, info)

    elif newpacket.Command == commands.Files():
        # supplier sent us list of files
        # handled by service_backups()
        Files(newpacket, info)

    elif newpacket.Command == commands.DeleteFile():
        # handled by service_supplier()
        DeleteFile(newpacket)

    elif newpacket.Command == commands.DeleteBackup():
        # handled by service_supplier()
        DeleteBackup(newpacket)

    elif newpacket.Command == commands.Correspondent():
        # TODO: contact asking for our current identity, not implemented yet
        Correspondent(newpacket)

    elif newpacket.Command == commands.Broadcast():
        # handled by service_broadcasting()
        Broadcast(newpacket, info)

    elif newpacket.Command == commands.Coin():
        # handled by service_accountant()
        Coin(newpacket, info)

    elif newpacket.Command == commands.RetrieveCoin():
        # handled by service_accountant()
        RetrieveCoin(newpacket, info)

    elif newpacket.Command == commands.Key():
        # handled by service_keys_registry()
        Key(newpacket, info)

    elif newpacket.Command == commands.AuditKey():
        # handled by service_keys_registry()
        AuditKey(newpacket, info)

    elif newpacket.Command == commands.Event():
        # handled by service_p2p_hookups()
        Event(newpacket, info)

    elif newpacket.Command == commands.Message():
        # handled by service_private_messages()
        Message(newpacket, info)

    elif newpacket.Command == commands.Contacts():
        # handled by service_customer_family()
        Contacts(newpacket, info)

    else:
        lg.warn('unexpected command received: %r from %r' % (newpacket, info))

    return False


def outbox(outpacket, wide, callbacks, target=None, route=None, response_timeout=None, keep_alive=True):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.outbox [%s:%s] to %s with route %r' % (outpacket.Command, outpacket.PacketID, nameurl.GetName(outpacket.RemoteID), route))
    return None


#------------------------------------------------------------------------------


def Ack(newpacket, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Ack %s from [%s] at %s://%s with %d bytes payload' % (newpacket.PacketID, nameurl.GetName(newpacket.CreatorID), info.proto, info.host, len(newpacket.Payload)))


def SendAck(packettoack, response='', wide=False, callbacks={}, remote_idurl=None):
    if remote_idurl is None:
        remote_idurl = packettoack.OwnerID
    result = signed.Packet(
        Command=commands.Ack(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packettoack.PacketID,
        Payload=response,
        RemoteID=remote_idurl,
    )
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendAck %s to %s  with %d bytes' % (result.PacketID, result.RemoteID, len(response)))
    gateway.outbox(result, wide=wide, callbacks=callbacks)
    return result


def SendAckNoRequest(remoteID, packetid, response='', wide=False, callbacks={}):
    result = signed.Packet(
        Command=commands.Ack(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packetid,
        Payload=response,
        RemoteID=remoteID,
    )
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendAckNoRequest packetID=%s to %s  with %d bytes' % (result.PacketID, result.RemoteID, len(response)))
    gateway.outbox(result, wide=wide, callbacks=callbacks)
    return result


#------------------------------------------------------------------------------


def Fail(newpacket):
    if _Debug:
        if newpacket.Payload:
            lg.warn('%r received from [%s] in %r' % (newpacket.Payload, newpacket.CreatorID, newpacket))
        else:
            lg.out(_DebugLevel, 'p2p_service.Fail from %s|%s packetID=%s : %s' % (newpacket.CreatorID, newpacket.OwnerID, newpacket.PacketID, newpacket.Payload))


def SendFail(request, response='', remote_idurl=None, wide=False):
    if remote_idurl is None:
        remote_idurl = request.OwnerID
    result = signed.Packet(
        Command=commands.Fail(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=request.PacketID,  # This is needed to identify Fail on remote side
        Payload=response,
        RemoteID=remote_idurl,
    )
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendFail packetID=%s to %s  with %d bytes' % (result.PacketID, result.RemoteID, len(response)))
    gateway.outbox(result, wide=wide)
    return result


def SendFailNoRequest(remoteID, packetID, response=''):
    result = signed.Packet(
        Command=commands.Fail(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packetID,
        Payload=response,
        RemoteID=remoteID,
    )
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendFailNoRequest packetID=%s to %s' % (result.PacketID, result.RemoteID))
    gateway.outbox(result)
    return result


#------------------------------------------------------------------------------


def Identity(newpacket, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Identity %s from [%s] at %s://%s with %d bytes payload' % (newpacket.PacketID, nameurl.GetName(newpacket.CreatorID), info.proto, info.host, len(newpacket.Payload)))


#     """
#     Normal node or Identity server is sending us a new copy of an identity for a contact of ours.
#     Checks that identity is signed correctly.
#     Sending requests to cache all sources (other identity servers) holding that identity.
#     """
#     # TODO:  move to service_gateway
#     newxml = newpacket.Payload
#     newidentity = identity.identity(xmlsrc=newxml)
#     # SECURITY
#     # check that identity is signed correctly
#     # old public key matches new one
#     # this is done in `UpdateAfterChecking()`
#     idurl = newidentity.getIDURL()
#     if not identitycache.HasKey(idurl):
#         lg.info('received new identity %s rev %r' % (idurl, newidentity.getRevisionValue()))
#     if not identitycache.UpdateAfterChecking(idurl, newxml):
#         lg.warn("ERROR has non-Valid identity")
#         return False
#     if my_id.isLocalIdentityReady():
#         if newidentity.getPublicKey() == my_id.getLocalIdentity().getPublicKey():
#             if newidentity.getRevisionValue() > my_id.getLocalIdentity().getRevisionValue():
#                 lg.warn('received my own identity from another user, but with higher revision number')
#                 reactor.callLater(0, my_id.rebuildLocalIdentity, new_revision=newidentity.getRevisionValue() + 1)  # @UndefinedVariable
#                 return False
#     latest_identity = id_url.get_latest_ident(newidentity.getPublicKey())
#     if latest_identity:
#         if latest_identity.getRevisionValue() > newidentity.getRevisionValue():
#             # check if received identity is the most recent revision number we every saw for that remote user
#             # in case we saw same identity with higher revision number need to reply with Fail packet and notify user
#             # this may happen after identity restore - the user starts counting revision number from 0
#             # but other nodes already store previous copies, user just need to jump to the most recent revision number
#             lg.warn('received new identity with out-dated revision number from %r' % idurl)
#             ident_packet = signed.Packet(
#                 Command=commands.Identity(),
#                 OwnerID=latest_identity.getIDURL(),
#                 CreatorID=latest_identity.getIDURL(),
#                 PacketID='identity:%s' % packetid.UniqueID(),
#                 Payload=latest_identity.serialize(),
#                 RemoteID=idurl,
#             )
#             reactor.callLater(0, packet_out.create, outpacket=ident_packet, wide=True, callbacks={}, keep_alive=False)  # @UndefinedVariable
#             return False
#     # Now that we have ID we can check the packet
#     if not newpacket.Valid():
#         # If not valid do nothing
#         lg.warn("not Valid packet from %s" % idurl)
#         return False
#     if not send_ack:
#         if _Debug:
#             lg.out(_DebugLevel, "p2p_service.Identity %s  idurl=%s  remoteID=%r  skip sending Ack()" % (
#                 newpacket.PacketID, idurl, newpacket.RemoteID))
#         return True
#     if newpacket.OwnerID == idurl:
#         if _Debug:
#             lg.out(_DebugLevel, "p2p_service.Identity %s  idurl=%s  remoteID=%r  sending wide Ack()" % (
#                 newpacket.PacketID, idurl, newpacket.RemoteID))
#     else:
#         if _Debug:
#             lg.out(_DebugLevel, "p2p_service.Identity %s  idurl=%s  remoteID=%r  but packet ownerID=%s   sending wide Ack()" % (
#                 newpacket.PacketID, idurl, newpacket.RemoteID, newpacket.OwnerID, ))
#     # wide=True : a small trick to respond to all known contacts of the remote user
#     reactor.callLater(0, SendAck, newpacket, wide=True)  # @UndefinedVariable
#     return True


def SendIdentity(remote_idurl, wide=False, timeout=None, callbacks={}):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    packet_id = 'identity:%s' % packetid.UniqueID()
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendIdentity to %s wide=%s packet_id=%r' % (nameurl.GetName(remote_idurl), wide, packet_id))
    result = signed.Packet(
        Command=commands.Identity(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=my_id.getLocalIdentity().serialize(),
        RemoteID=remote_idurl,
    )
    gateway.outbox(result, wide=wide, callbacks=callbacks, response_timeout=timeout)
    return result


#------------------------------------------------------------------------------


def RequestService(request, info):
    if _Debug:
        try:
            service_info = serialization.BytesToDict(request.Payload)
        except:
            lg.exc()
            service_info = {}
        lg.out(_DebugLevel, 'p2p_service.RequestService %s with "%s" in %d bytes' % (request.PacketID, service_info.get('name', 'unknown service name'), len(request.Payload)))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendRequestService(remote_idurl, service_name, json_payload={}, wide=False, callbacks={}, timeout=None, packet_id=None):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    service_info = {
        'name': service_name,
        'payload': json_payload,
    }
    service_info_raw = serialization.DictToBytes(service_info)
    if packet_id is None:
        packet_id = packetid.UniqueID()
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendRequestService "%s" to %s with %d bytes pid:%s timeout:%r' % (service_name, remote_idurl, len(service_info_raw), packet_id, timeout))
    result = signed.Packet(
        Command=commands.RequestService(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=service_info_raw,
        RemoteID=remote_idurl,
    )
    gateway.outbox(result, wide=wide, callbacks=callbacks, response_timeout=timeout)
    return result


def CancelService(request, info):
    if _Debug:
        try:
            service_info = serialization.BytesToDict(request.Payload)
        except:
            lg.exc()
            service_info = {}
        lg.out(_DebugLevel, 'p2p_service.CancelService %s with "%s" in %d bytes' % (request.PacketID, service_info.get('name', 'unknown service name'), len(request.Payload)))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendCancelService(remote_idurl, service_name, json_payload={}, wide=False, callbacks={}, timeout=None):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    service_info = {
        'name': service_name,
        'payload': json_payload,
    }
    service_info_raw = serialization.DictToBytes(service_info)
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendCancelService "%s" to %s with %d bytes payload' % (service_name, remote_idurl, len(service_info_raw)))
    result = signed.Packet(
        Command=commands.CancelService(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packetid.UniqueID(),
        Payload=service_info_raw,
        RemoteID=remote_idurl,
    )
    gateway.outbox(result, wide=wide, callbacks=callbacks, response_timeout=timeout)
    return result


#------------------------------------------------------------------------------


def ListFiles(request, info):
    """
    We will want to use this to see what needs to be resent, and expect normal
    case is very few missing.
    This is to build the ``Files()`` we are holding for a customer.
    You run service_list_files locally and send ListFiles() to your suppliers
    They repply fith Files() if service_supplier is started on their side
    """
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.ListFiles %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendListFiles(target_supplier, customer_idurl=None, key_id=None, query_items=[], wide=False, callbacks={}, timeout=None):
    """
    This is used as a request method from your supplier : if you send him a ListFiles() packet
    he will reply you with a list of stored files in a Files() packet.
    """
    if timeout is None:
        timeout = settings.P2PTimeOut()
    MyID = my_id.getIDURL()
    if not customer_idurl:
        customer_idurl = MyID
    if not str(target_supplier).isdigit():
        RemoteID = target_supplier
    else:
        RemoteID = contactsdb.supplier(target_supplier, customer_idurl=customer_idurl)
    if not RemoteID:
        lg.warn('RemoteID is empty target_supplier=%s' % str(target_supplier))
        return None
    if not key_id:
        # key_id = global_id.MakeGlobalID(idurl=customer_idurl, key_alias='customer')
        # TODO: due to issue with "customer" key backup/restore decided to always use my "master" key
        # to retrieve my list files info from supplier
        # expect remote user always poses my master public key from my identity.
        # probably require more work to build more reliable solution without using my master key at all
        # when my identity rotated supplier first needs to receive my new identity and then sending ListFiles()
        key_id = my_id.getGlobalID(key_alias='master')
    else:
        key_id = my_keys.latest_key_id(key_id)
    if not my_keys.is_key_registered(key_id) or not my_keys.is_key_private(key_id):
        lg.warn('key %r not exist or public, my "master" key to be used with ListFiles() packet' % key_id)
        key_id = my_id.getGlobalID(key_alias='master')
    PacketID = '%s:%s' % (
        key_id,
        packetid.UniqueID(),
    )
    if not query_items:
        query_items = ['*']
    Payload = serialization.DictToBytes({'items': query_items})
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendListFiles %r to %r of customer %r with query : %r' % (PacketID, nameurl.GetName(RemoteID), nameurl.GetName(customer_idurl), query_items))
    result = signed.Packet(
        Command=commands.ListFiles(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=PacketID,
        Payload=Payload,
        RemoteID=RemoteID,
    )
    gateway.outbox(result, wide=wide, callbacks=callbacks, response_timeout=timeout)
    return result


#------------------------------------------------------------------------------


def Files(request, info):
    """
    A directory list came in from some supplier or another customer.
    """
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Files %d bytes in [%s] from %s by %s|%s' % (len(request.Payload), request.PacketID, nameurl.GetName(request.RemoteID), nameurl.GetName(request.OwnerID), nameurl.GetName(request.CreatorID)))


def SendFiles(idurl, raw_list_files_info, packet_id, callbacks={}, timeout=None):
    """
    Sending information about known files stored locally for given customer (if you are supplier).
    You can also send a list of your files to another user if you wish to grand access.
    This will not send any personal data : only file names, ids, versions, etc.
    So pass list of files in encrypted form in the `payload` or leave it empty.
    """
    if timeout is None:
        timeout = settings.P2PTimeOut()
    MyID = my_id.getIDURL()
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendFiles %d bytes in packetID=%s' % (len(raw_list_files_info), packet_id))
        lg.out(_DebugLevel, '  to remoteID=%s' % idurl)
    newpacket = signed.Packet(
        Command=commands.Files(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=packet_id,
        Payload=raw_list_files_info,
        RemoteID=idurl,
    )
    result = gateway.outbox(newpacket, callbacks=callbacks, response_timeout=timeout)
    return result


#------------------------------------------------------------------------------


def Data(request):
    """
    This is when we 1) save my requested data to restore the backup 2) or save
    the customer file on our local HDD.
    """
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Data %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendData(raw_data, ownerID, creatorID, remoteID, packetID, callbacks={}):
    newpacket = signed.Packet(
        Command=commands.Data(),
        OwnerID=ownerID,
        CreatorID=creatorID,
        PacketID=packetID,
        Payload=raw_data,
        RemoteID=remoteID,
    )
    result = gateway.outbox(newpacket, callbacks=callbacks)
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendData %d bytes in packetID=%s' % (len(raw_data), packetID))
        lg.out(_DebugLevel, '  to remoteID=%s  ownerID=%s  creatorID=%s' % (remoteID, ownerID, creatorID))
    return newpacket, result


def Retrieve(request):
    """
    Customer is asking us for data he previously stored with us.

    We send with ``outboxNoAck()`` method because he will ask again if
    he does not get it
    """
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Retrieve %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendRetreive(ownerID, creatorID, packetID, remoteID, payload='', response_timeout=None, callbacks={}):
    newpacket = signed.Packet(
        Command=commands.Retrieve(),
        OwnerID=ownerID,
        CreatorID=creatorID,
        PacketID=packetID,
        Payload=payload,
        RemoteID=remoteID,
    )
    result = gateway.outbox(newpacket, callbacks=callbacks, response_timeout=response_timeout)
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendRetreive packetID=%s' % packetID)
        lg.out(_DebugLevel, '  remoteID=%s  ownerID=%s  creatorID=%s' % (remoteID, ownerID, creatorID))
    return result


#------------------------------------------------------------------------------


def DeleteFile(request):
    """
    Delete one ore multiple files (that belongs to another user) or folders on my machine.
    """
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.DeleteFile %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendDeleteFile(SupplierID, pathID):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendDeleteFile SupplierID=%s PathID=%s' % (SupplierID, pathID))
    MyID = my_id.getIDURL()
    PacketID = pathID
    RemoteID = SupplierID
    result = signed.Packet(
        Command=commands.DeleteFile(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=PacketID,
        Payload='',
        RemoteID=RemoteID,
    )
    gateway.outbox(result)
    return result


def SendDeleteListPaths(SupplierID, ListPathIDs):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendDeleteListPaths SupplierID=%s PathIDs: %s' % (SupplierID, ListPathIDs))
    MyID = my_id.getIDURL()
    PacketID = packetid.UniqueID()
    RemoteID = SupplierID
    Payload = '\n'.join(ListPathIDs)
    result = signed.Packet(
        Command=commands.DeleteFile(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=PacketID,
        Payload=Payload,
        RemoteID=RemoteID,
    )
    gateway.outbox(result)
    return result


#------------------------------------------------------------------------------


def DeleteBackup(request):
    """
    Delete one or multiple backups on my machine.
    """
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.DeleteBackup %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendDeleteBackup(SupplierID, BackupID):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendDeleteBackup SupplierID=%s  BackupID=%s' % (SupplierID, BackupID))
    MyID = my_id.getIDURL()
    PacketID = packetid.RemotePath(BackupID)
    RemoteID = SupplierID
    result = signed.Packet(
        Command=commands.DeleteBackup(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=PacketID,
        Payload='',
        RemoteID=RemoteID,
    )
    gateway.outbox(result)
    return result


def SendDeleteListBackups(SupplierID, ListBackupIDs):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendDeleteListBackups SupplierID=%s BackupIDs: %s' % (SupplierID, ListBackupIDs))
    MyID = my_id.getIDURL()
    PacketID = packetid.UniqueID()
    RemoteID = SupplierID
    Payload = '\n'.join(ListBackupIDs)
    result = signed.Packet(
        Command=commands.DeleteBackup(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=PacketID,
        Payload=Payload,
        RemoteID=RemoteID,
    )
    gateway.outbox(result)
    return result


#------------------------------------------------------------------------------


def Correspondent(request):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Correspondent %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))
    # TODO: need to connect users here


#------------------------------------------------------------------------------


def RequestDeleteListBackups(backupIDs):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.RequestDeleteListBackups wish to delete %d backups' % len(backupIDs))
    customers = {}
    for backupID in backupIDs:
        customer_idurl = packetid.CustomerIDURL(backupID)
        if customer_idurl not in customers:
            customers[customer_idurl] = set()
        customers[customer_idurl].add(backupID)
    for customer_idurl, backupID_set in customers.items():
        for supplier_idurl in contactsdb.suppliers(customer_idurl=customer_idurl):
            if supplier_idurl:
                SendDeleteListBackups(supplier_idurl, list(backupID_set))


def RequestDeleteListPaths(pathIDs):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.RequestDeleteListPaths wish to delete %d paths' % len(pathIDs))
    customers = {}
    for pathID in pathIDs:
        customer_idurl = packetid.CustomerIDURL(pathID)
        if customer_idurl not in customers:
            customers[customer_idurl] = set()
        customers[customer_idurl].add(pathID)
    for customer_idurl, pathID_set in customers.items():
        for supplier_idurl in contactsdb.suppliers(customer_idurl=customer_idurl):
            if supplier_idurl:
                SendDeleteListPaths(supplier_idurl, list(pathID_set))


def CheckWholeBackup(BackupID):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.CheckWholeBackup with BackupID=' + BackupID)


#-------------------------------------------------------------------------------


def Broadcast(request, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Broadcast %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s  sender_idurl=%s' % (request.RemoteID, request.OwnerID, request.CreatorID, info.sender_idurl))


def SendBroadcastMessage(outpacket):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendBroadcastMessage to %s' % outpacket.RemoteID)
    gateway.outbox(outpacket)
    return outpacket


#------------------------------------------------------------------------------


def Coin(request, info):
    if _Debug:
        try:
            input_coins = serialization.BytesToDict(request.Payload, keys_to_text=True)
        except:
            lg.exc()
            input_coins = []
        lg.out(_DebugLevel, 'p2p_service.Coin from %s with %d coins' % (nameurl.GetName(info.sender_idurl), len(input_coins)))


def SendCoin(remote_idurl, coins, packet_id=None, wide=False, callbacks={}):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendCoin to %s with %d records' % (remote_idurl, len(coins)))
    if packet_id is None:
        packet_id = packetid.UniqueID()
    outpacket = signed.Packet(
        Command=commands.Coin(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=serialization.DictToBytes(coins, keys_to_text=True),
        RemoteID=remote_idurl,
    )
    gateway.outbox(outpacket, wide=wide, callbacks=callbacks)
    return outpacket


def RetrieveCoin(request, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.RetrieveCoin from %s : %s' % (nameurl.GetName(info.sender_idurl), request.Payload))


def SendRetrieveCoin(remote_idurl, query, wide=False, callbacks={}):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendRetrieveCoin to %s' % remote_idurl)
    outpacket = signed.Packet(
        Command=commands.RetrieveCoin(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packetid.UniqueID(),
        Payload=serialization.DictToBytes(query),
        RemoteID=remote_idurl,
    )
    gateway.outbox(outpacket, wide=wide, callbacks=callbacks)
    return outpacket


#------------------------------------------------------------------------------


def Key(request, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Key %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from senderID=%s to remoteID=%s  ownerID=%s  creatorID=%s  ' % (info.sender_idurl, request.RemoteID, request.OwnerID, request.CreatorID))


def SendKey(
    remote_idurl,
    encrypted_key_data,
    packet_id=None,
    wide=False,
    callbacks={},
    timeout=None,
):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    if packet_id is None:
        packet_id = packetid.UniqueID()
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendKey [%s] to %s with %d bytes encrypted key data' % (packet_id, remote_idurl, len(encrypted_key_data)))
    outpacket = signed.Packet(
        Command=commands.Key(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=encrypted_key_data,
        RemoteID=remote_idurl,
    )
    gateway.outbox(outpacket, wide=wide, callbacks=callbacks, response_timeout=timeout)
    return outpacket


def AuditKey(request, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.AuditKey %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s  sender_idurl=%s' % (request.RemoteID, request.OwnerID, request.CreatorID, info.sender_idurl))


def SendAuditKey(remote_idurl, encrypted_payload, packet_id=None, timeout=None, wide=False, callbacks={}):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendAuditKey to %s with %d bytes in json payload data' % (remote_idurl, len(encrypted_payload)))
    if timeout is None:
        timeout = settings.P2PTimeOut()
    if packet_id is None:
        packet_id = packetid.UniqueID()
    outpacket = signed.Packet(
        Command=commands.AuditKey(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=encrypted_payload,
        RemoteID=remote_idurl,
    )
    gateway.outbox(outpacket, wide=wide, callbacks=callbacks, response_timeout=timeout)
    return outpacket


#------------------------------------------------------------------------------


def Event(request, info):
    if _Debug:
        try:
            e_json = serialization.BytesToDict(request.Payload, keys_to_text=True)
            e_json['event_id']
            e_json['payload']
        except:
            lg.exc()
            return
        lg.out(_DebugLevel, 'p2p_service.Event %s from %s with %d bytes in json' % (e_json['event_id'], info.sender_idurl, len(request.Payload)))


def SendEvent(remote_idurl, event_id, payload=None, producer_id=None, consumer_id=None, queue_id=None, message_id=None, created=None, packet_id=None, wide=False, callbacks={}, response_timeout=None):
    if response_timeout is None:
        response_timeout = settings.P2PTimeOut()
    if packet_id is None:
        packet_id = packetid.UniqueID()
    e_json = {
        'event_id': event_id,
        'payload': payload,
    }
    if producer_id is not None:
        e_json['producer_id'] = producer_id
    if consumer_id is not None:
        e_json['consumer_id'] = consumer_id
    if queue_id is not None:
        e_json['queue_id'] = queue_id
    if message_id is not None:
        e_json['message_id'] = message_id
    if created is not None:
        e_json['created'] = created
    e_json_src = serialization.DictToBytes(e_json)
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendEvent to %s with %d bytes message json data' % (remote_idurl, len(e_json_src)))
    outpacket = signed.Packet(
        Command=commands.Event(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=e_json_src,
        RemoteID=remote_idurl,
    )
    gateway.outbox(outpacket, wide=wide, callbacks=callbacks, response_timeout=response_timeout)
    return outpacket


#------------------------------------------------------------------------------


def Message(request, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Message %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendMessage(remote_idurl, packet_id=None, payload=None, wide=True, callbacks={}, response_timeout=None):
    if packet_id is None:
        packet_id = packetid.UniqueID()
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendMessage to %s with packet_id=%s' % (nameurl.GetName(remote_idurl), packet_id))
    outpacket = signed.Packet(
        Command=commands.Message(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=packet_id,
        Payload=payload,
        RemoteID=remote_idurl,
    )
    result = gateway.outbox(outpacket, wide=wide, callbacks=callbacks, response_timeout=response_timeout)
    return result, outpacket


#------------------------------------------------------------------------------


def Contacts(request, info):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.Contacts %d bytes in [%s]' % (len(request.Payload), request.PacketID))
        lg.out(_DebugLevel, '  from remoteID=%s  ownerID=%s  creatorID=%s' % (request.RemoteID, request.OwnerID, request.CreatorID))


def SendContacts(remote_idurl, json_payload={}, wide=False, callbacks={}):
    MyID = my_id.getIDURL()
    if _Debug:
        lg.out(_DebugLevel, 'p2p_service.SendContacts to %s' % nameurl.GetName(remote_idurl))
    PacketID = packetid.UniqueID()
    try:
        json_payload['type']
        json_payload['space']
    except:
        lg.err()
        return None
    Payload = serialization.DictToBytes(json_payload)
    result = signed.Packet(
        Command=commands.Contacts(),
        OwnerID=MyID,
        CreatorID=MyID,
        PacketID=PacketID,
        Payload=Payload,
        RemoteID=remote_idurl,
    )
    gateway.outbox(result, wide=wide, callbacks=callbacks)
    return result


#------------------------------------------------------------------------------



#!/usr/bin/python
# identity.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (identity.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: identity.

.. role:: red

This is a core module.

Identity might better be called business card, since it is really
mostly just your contact info. But more than that since public key.
Hum.

The whole idea of identity files is that they can be stored anywhere.
User can decide to put his identity file on his own host, or he can use some another trusted public host.
All identity files is signed with user's Key and client code should verify signatures.
Even if someone tried to fake my identity - this will be refused by other nodes in the network.

Example in http://seed.bitdust.io/veselin.xml

Could have info just be XML on a web page.  So an Identity could be a URL.
If we do this, then we get the scaling of DNS for free.

BitDust could have mappings from unique short names to URLs for
identities.  Or we could just make sure that the MD5 of the URL was
always unique (this is 16 bytes).  Or we could go with the primary URL.
Hum. BitDust does need some sort of unique identifier, as do others.
Also, would be nice to be able to send someone a list of their
current 64 nodes they use, or the 300 that use them, without
crazy amounts of data.  For display
of backup status I would like a unique identifier that is short,
like 15 chars or something.  Can get main part of URLs that is
short like that, say "id.ai/vince1".  Seems funny to limit the
length of a URL like this, but we could as there will be
identity servers and they can get very short names like id.ai.
If we forced them all to be ".com/" we could display "id:vince1",
or "cate:vince1", which is short.  The force could just be
a lower rating for those that did not do this.  Could have
short name take up 2 lines out of my 5, and just plan on
letting user click on things to change what things are
displayed.  So they have 5 lines and full URL takes 2,
but last of URL is just 1 line, and maybe they don't
care what the name is and just display other stuff.

If we open it up a bit, people could just use tinyurl.com/foolksd
if their URL was too long.  We could limit primary URL
to 32 chars and not really have any trouble.  Can say it
must start with http:// (so we don't have to store that)
and be less than 32 chars after that.  If does not really
start with http:// then tiny URL can fix that too.

It is a bit like bittorrent using a .torrent file on the
web to get the original start for things.  This seems good.

All Identity/business-card info in XML:
PublicKey
primary URL for this identity (say https://cate.com/vinceID)
backup URLs for this identity (maybe URLs) that secondary this identity
contact: domain/port  pairs  (good even if IP out of date)
contact: IP/port  pairs    (good even if DNS in trouble)
contact: nat-traverser/ID pairs  (for those without fixed IPs)
contact: emails for this identity (recommend people have one just in case)
contact: http://foobar.com/data.pl    (use a web account to forward requests/data)
scrubbers: URL1, URL2, URL3           (people who can access our data if we stop checking it)
date
version number
revision number
signature on all identity info

Contact list has enough info we can tell what protocol to use.
User could put in order he prefers us to try the contact methods.
So we might have a list like:
    tcp://123.45.67.89:9876
    udp://user123@idserver456.com
    http://123.45.67.89:8765

Really best if all the identity servers use SSL.
We could make certificates for identity servers, but might
not bother as it is probably best if they have ones that
web browsers understand.
On the other hand, this might be a good way to get into the certificate-authority business. :-)

Having XML on a web site means we can start before we
really have identity server code.  Also means it is
easy to grow and debug.

As long as identity server is just some CGI that we can run on
any ISP anywhere (pair.com etc), and users always use 3 identity servers,
then we could just pop up as many as needed on ISPs all over
the world as fast as we grew.  It would not be an infrastructure
limitation really.  And we don't need to farm it out.
Identity's are signed and have a revision number, so an identity
server can pass on an update that it gets to other servers listed
for that identity (in case there is network partition funnyness)
to help keep all of them updated.

But we really want actually people to run own identity servers and not rely on us.
ID server is a first target to external hacker to attack actually if he wants to brake/block the network.
Also ownership is important - 1000 servers owned by one company can be easily attacked via "legal/court" order.
But 1000 servers owned and maintained by 1000 people in 100 different countries is completely different scenario.
So more distribution of ID servers we will have in BitDust - more secured and protected network will become.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from six.moves import range  # @UnresolvedImport

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import sys

from xml.dom import minidom, Node
from xml.dom.minidom import getDOMImplementation

#------------------------------------------------------------------------------

try:
    from bitdust.logs import lg
except:
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    from bitdust.logs import lg

#------------------------------------------------------------------------------

from bitdust.main import settings

from bitdust.system import bpio

from bitdust.lib import strng
from bitdust.lib import nameurl

from bitdust.crypt import key

from bitdust.userid import global_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

default_identity_src = """<?xml version="1.0" encoding="ISO-8859-1"?>
<identity>
  <sources></sources>
  <contacts></contacts>
  <certificates></certificates>
  <scrubbers></scrubbers>
  <postage>1</postage>
  <date></date>
  <version></version>
  <revision>0</revision>
  <publickey></publickey>
  <signature></signature>
</identity>"""

#------------------------------------------------------------------------------


class identity(object):

    """
    We are passed an XML version of an identity and make an Identity. Also can
    construct an Identity by providing all fields. The fields is:

    * sources : list of URLs, first is primary URL and name
    * contacts : list of ways to contact this identity
    * certificates : signatures by identity servers
    * scrubbers : list of URLs for people allowed to scrub
    * postage : a price for message delivery if not on correspondents list
    * date : the date an time when that identity was created
    * version : a version string - some info about BitDust software and platform
    * revision : every time identity were modified this value will be increased by 1
    * publickey : the public part of user's key, string in twisted.conch.ssh format
    * signature : digital signature to protect the file
    """
    sources = []  # list of URLs, first is primary URL and name
    contacts = []  # list of ways to contact this identity
    certificates = []  # signatures by identity servers
    scrubbers = []  # list of URLs for people allowed to scrub
    postage = b'1'  # a price for message delivery if not on correspondents list
    date = b''  # date
    version = b''  # version string
    revision = b'0'  # revision number, every time my id were modified this value will be increased by 1
    publickey = b''  # string in twisted.conch.ssh format
    signature = b''  # digital signature

    def __init__(self, sources=[], contacts=[], certificates=[], scrubbers=[], postage=b'1', date=b'', version=b'', revision=b'0', publickey=b'', xmlsrc=None, filename=b''):

        self.sources = []
        if sources:
            self.setSources(sources)
        self.contacts = contacts
        self.certificates = certificates
        self.scrubbers = scrubbers
        self.postage = postage
        self.date = date
        self.version = version
        self.revision = revision
        self.publickey = publickey

        if publickey:
            self.sign()
        else:
            self.signature = b''
            # no point in signing if no public key listed, probably about to unserialize something

        if xmlsrc is not None:
            self.unserialize(xmlsrc)

        if filename:
            self.unserialize(bpio.ReadTextFile(filename))

        if not xmlsrc and not filename:
            self.default()

    def __repr__(self, *args, **kwargs):
        return object.__repr__(self, *args, **kwargs) + '[%s]' % self.makehash(hexdigest=True)

    def clear_data(self):
        """
        Erase all fields data, clear identity.
        """
        self.sources = []  # list of URLs for fetching this identiy, first is primary URL and name - called IDURL
        self.certificates = []  # identity servers each sign the source they are with - hash just (IDURL + publickey)
        self.publickey = b''  # string
        self.contacts = []  # list of ways to contact this identity
        self.scrubbers = []  # list of URLs for people allowed to scrub
        self.date = b''  # date
        self.postage = b'1'  # postage price for message delivery if not on correspondents list
        self.version = b''  # version string
        self.signature = b''  # digital signature
        self.revision = b'0'  # revision number

    def default(self):
        """
        Set a "blank" identity.
        """
        global default_identity_src
        self.unserialize(default_identity_src)

    #------------------------------------------------------------------------------

    def isCorrect(self):
        """
        Do some checking on the object fields.
        """
        if len(self.contacts) == 0:
            return False
        if len(self.sources) == 0:
            return False
        if not self.publickey:
            return False
        if not self.signature:
            return False
        if not self.revision:
            return False
        original_sources = self.getSources(as_originals=True)
        if len(set(original_sources)) != len(self.sources):
            lg.warn('original identity sources are duplicated: %r' % original_sources)
            return False
        if len(original_sources) > settings.MaximumIdentitySources():
            lg.warn('too much sources')
            return False
        if len(original_sources) < settings.MinimumIdentitySources():
            lg.warn('too few sources')
            return False
        try:
            int(self.revision)
        except:
            lg.warn('identity revision: %s' % self.revision)
            return False
        names = set()
        for source in original_sources:
            if not source:
                lg.warn('found empty source')
                return False
            proto, host, port, filename = nameurl.UrlParse(source)
            if filename.count('/'):
                lg.warn('incorrect identity name: %s' % filename)
                return False
            name, justxml = filename.split('.')
            names.add(name)
            # SECURITY check that name is simple
            if justxml != 'xml':
                lg.warn('incorrect identity name: %s' % filename)
                return False
            if len(name) > settings.MaximumUsernameLength():
                lg.warn('incorrect identity name: %s' % filename)
                return False
            if len(name) < settings.MinimumUsernameLength():
                lg.warn('incorrect identity name: %s' % filename)
                return False
            for c in name:
                if c not in settings.LegalUsernameChars():
                    lg.warn('incorrect identity name: %s' % filename)
                    return False
        if len(names) > 1:
            lg.warn('names are not consistent: %s' % str(names))
            return False
        return True

    def makehash(self, hexdigest=False):
        """
        http://docs.python.org/lib/module-urlparse.html Note that certificates
        and signatures are not part of what is hashed. PREPRO Thinking of
        standard that fields have labels and empty fields are left out,
        including label, so future versions could have same signatures as older
        which had fewer fields - can just do this for fields after these, so
        maybe don't need to change anything for now.
        Don't include certificate - so identity server can just add it.
        """
        sep = b'-'
        hsh = b''
        hsh += sep + sep.join(map(lambda s: s.original(), self.sources))
        hsh += sep + sep.join(self.contacts)
        # hsh += sep + sep.join(self.certificates)
        hsh += sep + sep.join(self.scrubbers)
        hsh += sep + self.postage
        hsh += sep + self.date.replace(b' ', b'_')
        hsh += sep + self.version
        hsh += sep + self.revision
        # lg.out(12, "identity.makehash: %r" % hsh)
        hashcode = key.Hash(hsh, hexdigest=hexdigest)
        return hashcode

    def sign(self):
        """
        Make a hash, generate digital signature on it and remember the
        signature.
        """
        hashcode = self.makehash()
        self.signature = key.Sign(hashcode)
        if not self.Valid():
            lg.warn('identity is not valid after making sign!')
            raise Exception('identity sign fails')

    def Valid(self):
        """
        This will make a hash and verify the signature by public key.

        PREPRO - should test certificate too.
        """
        # print('Valid %r' % self.signature)
        hashcode = self.makehash()
        result = key.VerifySignature(
            self.publickey,
            hashcode,
            self.signature,
        )
        return result

    #------------------------------------------------------------------------------

    def encrypt(self, inp):
        """
        Encrypt some `inp` data with public key from this identity object.
        You can use this method to send a private messages addressed to the owner of this identity.
        """
        return key.EncryptOpenSSHPublicKey(self.publickey, inp)

    def decrypt(self, inp, private_key=None):
        """
        Decrypt `inp` data with private key provided (callable, key_id or openssh format).
        When `private_key` is not provided will decrypt using my own "master" key.
        """
        if private_key is None:
            return key.DecryptLocalPrivateKey(inp)
        if callable(private_key):
            return private_key(inp)
        from bitdust.crypt import my_keys
        if my_keys.is_valid_key_id(private_key):
            return my_keys.decrypt(private_key, inp)
        return key.DecryptOpenSSHPrivateKey(private_key, inp)

    #------------------------------------------------------------------------------

    def unserialize(self, xmlsrc):
        """
        A smart method to load object fields data from XML content.
        """
        try:
            doc = minidom.parseString(strng.to_bin(xmlsrc))
        except:
            lg.warn('failed reading identity source from %d bytes' % len(strng.to_bin(xmlsrc)))
            return
        self.clear_data()
        self.from_xmlobj(doc.documentElement)

    def unserialize_object(self, xmlobject):
        """
        This is almost same but load data from existing DOM object.
        """
        self.clear_data()
        self.from_xmlobj(xmlobject)

    def unserialize_json(self, json_data):
        """
        This loads data from Python dictionary.
        """
        self.clear_data()
        self.setSources(json_data['sources'])
        self.setContacts(json_data['contacts'])
        self.setCertificates(json_data['certificates'])
        self.setScrubbers(json_data['scrubbers'])
        self.setDate(json_data['date'])
        self.setPostage(json_data['postage'])
        self.setVersion(json_data['version'])
        self.setRevision(json_data['version'])
        self.setPublicKey(json_data['publickey'])
        self.setSignature(json_data['signature'])

    def serialize(self, as_text=False):
        """
        A method to generate XML content for that identity object.

        Used to save identity on disk or transfer over network.
        """
        if as_text:
            return strng.to_text(self.toxml()[0].strip())
        return self.toxml()[0].strip()

    def serialize_object(self):
        """
        Almost the same but return a DOM object.
        """
        return self.toxml()[1]

    def serialize_json(self):
        """
        A method to generate JSON dictionary for that identity object.

        Used to represent identity in API methods.
        """
        return {
            'name': self.getIDName(),
            'idurl': strng.to_text(self.getIDURL().original()),
            'global_id': global_id.MakeGlobalID(idurl=self.getIDURL().original()),
            'sources': [strng.to_text(i) for i in self.getSources(as_originals=True)],
            'contacts': [strng.to_text(i) for i in self.getContacts()],
            'certificates': [strng.to_text(i) for i in self.certificates],
            'scrubbers': [strng.to_text(i) for i in self.scrubbers],
            'postage': strng.to_text(self.postage),
            'date': strng.to_text(self.date),
            'version': strng.to_text(self.version),
            'revision': strng.to_text(self.revision),
            'publickey': strng.to_text(self.publickey),
            'signature': strng.to_text(self.signature),
        }

    def toxml(self):
        """
        Call this to convert to XML format.
        """
        impl = getDOMImplementation()

        doc = impl.createDocument(None, 'identity', None)
        root = doc.documentElement

        sources = doc.createElement('sources')
        root.appendChild(sources)
        for source in self.sources:
            n = doc.createElement('source')
            n.appendChild(doc.createTextNode(strng.to_text(source.original())))
            sources.appendChild(n)

        contacts = doc.createElement('contacts')
        root.appendChild(contacts)
        for contact in self.contacts:
            n = doc.createElement('contact')
            n.appendChild(doc.createTextNode(strng.to_text(contact)))
            contacts.appendChild(n)

        certificates = doc.createElement('certificates')
        root.appendChild(certificates)
        for certificate in self.certificates:
            n = doc.createElement('certificate')
            n.appendChild(doc.createTextNode(strng.to_text(certificate)))
            certificates.appendChild(n)

        scrubbers = doc.createElement('scrubbers')
        root.appendChild(scrubbers)
        for scrubber in self.scrubbers:
            n = doc.createElement('scrubber')
            n.appendChild(doc.createTextNode(strng.to_text(scrubber)))
            scrubbers.appendChild(n)

        postage = doc.createElement('postage')
        postage.appendChild(doc.createTextNode(strng.to_text(self.postage)))
        root.appendChild(postage)

        date = doc.createElement('date')
        date.appendChild(doc.createTextNode(strng.to_text(self.date)))
        root.appendChild(date)

        version = doc.createElement('version')
        version.appendChild(doc.createTextNode(strng.to_text(self.version)))
        root.appendChild(version)

        revision = doc.createElement('revision')
        revision.appendChild(doc.createTextNode(strng.to_text(self.revision)))
        root.appendChild(revision)

        publickey = doc.createElement('publickey')
        publickey.appendChild(doc.createTextNode(strng.to_text(self.publickey)))
        root.appendChild(publickey)

        signature = doc.createElement('signature')
        signature.appendChild(doc.createTextNode(strng.to_text(self.signature)))
        root.appendChild(signature)

        xmlsrc = doc.toprettyxml(indent='  ', newl='\n', encoding='utf-8')
        return xmlsrc, root, doc

    def from_xmlobj(self, root_node):
        """
        This is to load identity fields from DOM object - used during ``unserialize`` procedure.
        """
        for xsection in root_node.childNodes:
            if xsection.nodeType != Node.ELEMENT_NODE:
                continue
            if xsection.tagName == 'sources':
                for xsources in xsection.childNodes:
                    for xsource in xsources.childNodes:
                        if (xsource.nodeType == Node.TEXT_NODE):
                            self.sources.append(id_url.ID_URL_FIELD(xsource.wholeText.strip()))
                            break
            elif xsection.tagName == 'contacts':
                for xcontacts in xsection.childNodes:
                    for xcontact in xcontacts.childNodes:
                        if (xcontact.nodeType == Node.TEXT_NODE):
                            self.contacts.append(strng.to_bin(xcontact.wholeText.strip()))
                            break
            elif xsection.tagName == 'certificates':
                for xcertificates in xsection.childNodes:
                    for xcertificate in xcertificates.childNodes:
                        if (xcertificate.nodeType == Node.TEXT_NODE):
                            self.certificates.append(strng.to_bin(xcertificate.wholeText.strip()))
                            break
            elif xsection.tagName == 'scrubbers':
                for xscrubbers in xsection.childNodes:
                    for xscrubber in xscrubbers.childNodes:
                        if (xscrubber.nodeType == Node.TEXT_NODE):
                            self.scrubbers.append(strng.to_bin(xscrubber.wholeText.strip()))
                            break
            elif xsection.tagName == 'postage':
                for xpostage in xsection.childNodes:
                    if (xpostage.nodeType == Node.TEXT_NODE):
                        self.postage = strng.to_bin(xpostage.wholeText.strip())
                        break
            elif xsection.tagName == 'date':
                for xkey in xsection.childNodes:
                    if (xkey.nodeType == Node.TEXT_NODE):
                        self.date = strng.to_bin(xkey.wholeText.strip())
                        break
            elif xsection.tagName == 'version':
                for xkey in xsection.childNodes:
                    if (xkey.nodeType == Node.TEXT_NODE):
                        self.version = strng.to_bin(xkey.wholeText.strip())
                        break
            elif xsection.tagName == 'revision':
                for xkey in xsection.childNodes:
                    if (xkey.nodeType == Node.TEXT_NODE):
                        self.revision = strng.to_bin(xkey.wholeText.strip())
                        break
            elif xsection.tagName == 'publickey':
                for xkey in xsection.childNodes:
                    if (xkey.nodeType == Node.TEXT_NODE):
                        self.publickey = strng.to_bin(xkey.wholeText.strip())
                        break
            elif xsection.tagName == 'signature':
                for xkey in xsection.childNodes:
                    if (xkey.nodeType == Node.TEXT_NODE):
                        self.signature = strng.to_bin(xkey.wholeText.strip())
                        break

    #------------------------------------------------------------------------------

    def getSources(self, as_fields=True, as_originals=False):
        """
        Return identity sources.
        """
        if as_originals:
            return id_url.to_original_list(self.sources)
        if not as_fields:
            return id_url.to_bin_list(self.sources)
        return self.sources

    def getIDURL(self, index=0, as_field=True, as_original=False):
        """
        Return a source IDURL - this is a user ID.
        Must have at least one IDURL in the ``sources``.
        """
        result = self.sources[index]
        if as_original:
            return result.original()
        if not as_field:
            return result.to_bin()
        return result

    def getIDName(self, index=0):
        """
        Return an account name - this is just a user name taken from IDURL:
            "veselin" for "http://id.bitdust.io/veselin.xml"
        """
        protocol, host, port, filename = nameurl.UrlParse(self.getIDURL(index))
        return filename.strip()[0:-4]

    def getIDHost(self, index=0):
        """
        Return a server host name where that identity is stored:
        "id.bitdust.io" for "http://id.bitdust.io/veselin.xml".
        """
        protocol, host, port, filename = nameurl.UrlParse(self.getIDURL(index))
        if port:
            host += ':' + str(port)
        return host

    def getContacts(self):
        """
        Return identity contacts list.
        """
        return self.contacts

    def getContactsNumber(self):
        """
        Return identity contacts number.
        """
        return len(self.contacts)

    def getContact(self, index=0):
        """
        Return a contact with given ``index`` number or None.
        """
        try:
            return self.contacts[index]
        except:
            return None

    def getContactHost(self, index=0):
        """
        Get the host name part of the contact.
        """
        protocol, host, port, filename = nameurl.UrlParse(self.contacts[index])
        return host

    def getContactPort(self, index=0):
        """
        Get the port part of the contact.
        """
        protocol, host, port, filename = nameurl.UrlParse(self.contacts[index])
        return port

    def getContactParts(self, index=0):
        """
        Return tuple with 4 parts of the contact: (proto, host, port, filename)
        """
        return nameurl.UrlParse(self.contacts[index])

    def getProtoParts(self, proto):
        """
        See ``getProtoContact``, return a tuple for given ``proto``: (proto,
        host, port, filename)
        """
        contact = self.getProtoContact(proto)
        if contact is None:
            return None, None, None, None
        return nameurl.UrlParse(contact)

    def getProtoHost(self, proto, default=None):
        """
        See ``getProtoParts``, return a host of a contact with given ``proto``.
        """
        protocol, host, port, filename = self.getProtoParts(proto)
        if host is None:
            return default
        return host

    def getContactIndex(self, proto='', host='', contact=''):
        """
        Search a first contact with given conditions.
        """
        for i in range(0, len(self.contacts)):
            c = strng.to_bin(self.contacts[i])
            if proto:
                if c.find(strng.to_bin(proto) + b'://') == 0:
                    return i
            if host:
                if c.find(b'://' + strng.to_bin(host)) == 0:
                    return i
            if contact:
                if c == strng.to_bin(contact):
                    return i
        return -1

    def getProtoContact(self, proto):
        """
        Search for first found contact with given ``proto``.

        Return None if not found a contact.
        """
        for contact in self.contacts:
            if contact.startswith(strng.to_bin(proto) + b'://'):
                return contact
        return None

    def getProtoOrder(self):
        """
        Return a list of "proto" parts of all contacts. In other words return a
        list of all supported protocols.

        This keeps the order of the protos - this is a sort of priority of the transports.
        """
        orderL = []
        for c in self.contacts:
            proto, host, port, filename = nameurl.UrlParse(c)
            orderL.append(proto)
        return orderL

    def getContactsAsTuples(self, as_text=False):
        """
        """
        result = []
        for c in self.contacts:
            proto, host = c.split(b'://')
            if as_text:
                result.append((strng.to_text(proto), strng.to_text(host)))
            else:
                result.append((proto, host))
        return result

    def getContactsByProto(self):
        """
        Return a dictionary of all contacts where keys are protos.
        """
        d = {}
        for i in range(len(self.contacts)):
            proto, x, x, x = nameurl.UrlParse(self.contacts[i])
            d[proto] = self.contacts[i]
        return d

    def getContactProto(self, index):
        """
        Return a proto part of the contact at given position.
        """
        c = self.getContact(index)
        if c is None:
            return None
        return nameurl.UrlParse(c)[0]

    def getPublicKey(self):
        """
        Returns public key as binary string.
        """
        return self.publickey

    def getSignature(self):
        """
        Returns signature as binary string.
        """
        return self.signature

    def getIP(self, proto=None):
        """
        A smart way to get the IP address of the user.

        Check TCP proto if available and get host from contact.
        """
        if proto:
            host = self.getProtoHost(proto)
            if host:
                return host
        return self.getProtoHost('tcp')

    def getDateStr(self):
        """
        Returns identity creation date as string.
        """
        return strng.to_text(self.date)

    def getPostageValue(self):
        """
        Returns price for message delivery if not on correspondents list.
        Not used at the moment.
        """
        return int(self.postage)

    def getVersionStr(self):
        """
        Returns a version string - some info about BitDust software and platform.
        Just informative thing, no real logic around it yet.
        """
        return strng.to_text(self.version)

    def getRevisionValue(self):
        """
        Returns identity revision number.
        Every time identity changes, identity must increment by one.
        """
        return int(self.revision)

    #------------------------------------------------------------------------------

    def setSources(self, sources_list):
        """
        """
        self.sources = []
        for source in sources_list:
            s = id_url.field(source)
            s.refresh()
            self.sources.append(s)

    def setContacts(self, contacts_list):
        """
        """
        self.contacts = []
        for cont in contacts_list:
            self.contacts.append(strng.to_bin(cont))

    def setContactsFromDict(self, contacts_dict, contacts_order=None):
        """
        """
        if contacts_order is None:
            contacts_order = list(contacts_dict.keys())
        for proto in contacts_order:
            self.contacts.append(strng.to_bin(contacts_dict[proto]))

    def setContact(self, contact, index):
        """
        Set a string value ``contact`` at given ``index`` position in the list.
        """
        try:
            self.contacts[index] = strng.to_bin(contact)
        except:
            lg.exc()

    def setProtoContact(self, proto, contact):
        """
        Found a contact with given ``proto`` and set its value or append a new
        contact.
        """
        for i in range(0, len(self.contacts)):
            proto_, _, _, _ = nameurl.UrlParse(self.contacts[i])
            if proto_.strip() == strng.to_bin(proto).strip():
                self.contacts[i] = strng.to_bin(contact)
                return
        self.contacts.append(strng.to_bin(contact))

    def setContactParts(self, index, protocol, host, port, filename):
        """
        Set a contact at given position by its 4 parts.
        """
        url = nameurl.UrlMake(protocol, host, port, filename)
        self.contacts[index] = strng.to_bin(url).strip()

    def setContactHost(self, host, index):
        """
        This is to set only host part of the contact.
        """
        protocol, _, port, filename = nameurl.UrlParse(self.contacts[index])
        url = nameurl.UrlMake(protocol, host, port, filename)
        self.contacts[index] = strng.to_bin(url).strip()

    def setContactPort(self, index, newport):
        """
        This is useful when listening port get changed.
        """
        protocol, host, _, filename = nameurl.UrlParse(self.contacts[index])
        url = nameurl.UrlMake(protocol, host, newport, filename)
        self.contacts[index] = strng.to_bin(url).strip()

    def setPublicKey(self, pub_key_raw):
        """
        """
        self.publickey = strng.to_bin(pub_key_raw)

    def setSignature(self, signature_raw):
        """
        """
        self.signature = strng.to_bin(signature_raw)

    def setCertificates(self, certificates_list):
        """
        Not used yet.
        """
        self.certificates = []
        for cert in certificates_list:
            self.certificates.append(strng.to_bin(cert))

    def setScrubbers(self, scrubbers_list):
        """
        Not used yet.
        """
        self.scrubbers = []
        for scrub in scrubbers_list:
            self.scrubbers.append(strng.to_bin(scrub))

    def setDate(self, date_string):
        """
        """
        self.date = strng.to_bin(date_string)

    def setPostage(self, postage_value):
        """
        """
        self.postage = strng.to_bin(str(postage_value))

    def setVersion(self, version_string):
        """
        """
        self.version = strng.to_bin(version_string).strip()

    def setRevision(self, revision):
        """
        """
        self.revision = strng.to_bin(str(revision))

    #------------------------------------------------------------------------------

    def clearContacts(self):
        """
        Erase all items in identity contacts list.
        """
        self.contacts = []

    def deleteProtoContact(self, proto):
        """
        Remove all contacts with given ``proto``.
        """
        for contact in self.contacts:
            if contact.find(strng.to_bin(proto) + b'://') == 0:
                self.contacts.remove(contact)

    def pushProtoContact(self, proto):
        """
        Move given protocol in the bottom of the contacts list.

        First contact in the list have more priority for remote machine,
        so we can manipulate our protos to get more p2p connections.
        Push less reliable protocols to the end of the list. This is to
        decrease its priority.
        """
        i = self.getContactIndex(proto=proto)
        if i < 0:
            return
        contact = self.contacts[i]
        del self.contacts[i]
        self.contacts.append(contact)

    def popProtoContact(self, proto):
        """
        Move given protocol to the top of the contacts list.

        This is to increase its priority.
        """
        i = self.getContactIndex(proto=proto)
        if i < 0:
            return
        contact = self.contacts[i]
        del self.contacts[i]
        self.contacts.insert(0, contact)


#-------------------------------------------------------------------------------


def test1():
    """
    Some tests.
    """
    from bitdust.userid import my_id
    myidentity = my_id.getLocalIdentity()
    print('getIP =', myidentity.getIP())
    if myidentity.Valid():
        print('myidentity is Valid!!!!')
    else:
        print('myidentity is not Valid')
        my_id.saveLocalIdentity()  # sign and save
        raise Exception('myidentity is not Valid')
    print('myidentity.contacts')
    print(myidentity.contacts)
    print('len myidentity.contacts ')
    print(len(myidentity.contacts))
    print('len myidentity.contacts[0] ')
    print(myidentity.contacts[0])
    con = myidentity.getContact()
    print('con:', con, type(con))
    protocol, machine, port, filename = nameurl.UrlParse(con)
    print(protocol, machine, port, filename)
    print('identity.main serialize:\n', myidentity.serialize())
    for index in range(myidentity.getContactsNumber()):
        proto, host, port, filename = myidentity.getContactParts(index)
        print('[%s] [%s] [%s] [%s]' % (proto, host, port, filename))


def test2():
    """
    More tests.
    """
    from bitdust.userid import my_id
    ident = my_id.buildDefaultIdentity()
    print(ident.serialize())


#------------------------------------------------------------------------------


def main():
    """
    This should print a current identity or create a new one.
    """
    settings.init()
    from bitdust.userid import my_id
    my_id.loadLocalIdentity()
    if my_id.isLocalIdentityReady():
        my_id.getLocalIdentity().sign()
        print(my_id.getLocalIdentity().serialize())
        print('Valid is: ', my_id.getLocalIdentity().Valid())
    else:
        if not key.InitMyKey():
            key.GenerateNewKey()
        ipaddr = '127.0.0.1'
        if len(sys.argv) > 2:
            ipaddr = sys.argv[2]
        rev = 0
        if len(sys.argv) > 3:
            rev = int(sys.argv[3])
        idurls = []
        if len(sys.argv) > 4:
            idurls = sys.argv[4:]
        my_id.setLocalIdentity(my_id.buildDefaultIdentity(name=sys.argv[1], ip=ipaddr, idurls=idurls, revision=rev))
        my_id.saveLocalIdentity()
        print(my_id.getLocalIdentity().serialize())
        print('Valid is: ', my_id.getLocalIdentity().Valid())
        my_id._LocalIdentity = None
        my_id.loadLocalIdentity()
    settings.shutdown()


def update():
    """
    A good way to check all things - load and sign again.
    Also will test rebuilding of the identity
    """
    from bitdust.userid import my_id
    bpio.init()
    settings.init()
    src = bpio.ReadTextFile(settings.LocalIdentityFilename())
    print(src)
    my_id.setLocalIdentity(identity(xmlsrc=src))
    my_id.getLocalIdentity().sign()
    my_id.saveLocalIdentity()
    print(my_id.getLocalIdentity().serialize())
    print(my_id.rebuildLocalIdentity(revision_up=True))
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    lg.set_debug_level(18)
    main()
    # update()



#!/usr/bin/python
# known_bismuth_nodes.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (known_bismuth_nodes.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_KnownBismuthNodes = None

#------------------------------------------------------------------------------


def default_nodes():
    """
    A set of Bismuth nodes currently maintained, see file `default_network.json` in the root folder.
    """
    from bitdust.main import network_config
    network_info = network_config.read_network_config_file()
    nodes = {}
    for node in network_info['service_bismuth_blockchain']['known_nodes']:
        nodes[node['host']] = {
            'host': node['host'],
            'node_tcp_port': node.get('node_tcp_port'),
            'mining_pool_tcp_port': node.get('mining_pool_tcp_port'),
            'explorer_http_port': node.get('explorer_http_port'),
        }
    return nodes


def by_host():
    global _KnownBismuthNodes

    if _KnownBismuthNodes is not None:
        return _KnownBismuthNodes

    _KnownBismuthNodes = default_nodes()
    return _KnownBismuthNodes


def nodes_by_host():
    return {node['host']: node['node_tcp_port'] for node in by_host().values() if node.get('node_tcp_port')}


def mining_pools_by_host():
    return {node['host']: node['mining_pool_tcp_port'] for node in by_host().values() if node.get('mining_pool_tcp_port')}


def explorers_by_host():
    return {node['host']: node['explorer_http_port'] for node in by_host().values() if node.get('explorer_http_port')}


def foundation_miners():
    from bitdust.main import network_config
    network_info = network_config.read_network_config_file()
    return network_info['service_bismuth_blockchain']['foundation_miners']



"""
Mempool module for Bismuth nodes
"""

import functools
# import base64
# import hashlib
import os
import sqlite3
import sys
import threading
from decimal import Decimal
import time

from polysign.signerfactory import SignerFactory

import essentials
from quantizer import quantize_two, quantize_eight

# from Cryptodome.Hash import SHA
# from Cryptodome.PublicKey import RSA
# from Cryptodome.Signature import PKCS1_v1_5

__version__ = '0.0.7e'
"""
0.0.5g - Add default param to mergedts for compatibility
0.0.5f - Using polysign
0.0.5e - add mergedts timestamp to tx for better handling of late txs
         quicker unfreeze
         less strict freezing
0.0.6b - Raise freeze tolerance to > 15 minutes old txs.
0.0.6c - Return last exception to client in all cases
0.0.7a - Add support for mandatory message addresses
0.0.7b - Reduce age of valid txns to 2 hours
0.0.7c - Remove unnecessary Decimal
0.0.7e - exclude too old txs from mempool balance, simplify mempool merge.
"""

DECIMAL0 = Decimal(0)

MEMPOOL = None

# If set to true, will always send empty Tx to other peers (but will accept theirs)
# Only to be used for debug/testing purposes
DEBUG_DO_NOT_SEND_TX = False

# Tx age limit (in seconds) - Default 82800
# REFUSE_OLDER_THAN = 82800
REFUSE_OLDER_THAN = 60*60*2  # reduced to 2 hours
# See also SQL_PURGE, SQL_MEMPOOL_GET and SQL_SELECT_ALL_VALID_TXS a few lines down.
# I used a filter on some requests rather than calling purge() every time.
# Maybe a systematic purge() would be easier and faster. To be tested.

# How long for freeze nodes that send late enough tx we already have in ledger
FREEZE_MIN = 5
"""
Common Sql requests
"""

# Create mempool table
SQL_CREATE = 'CREATE TABLE IF NOT EXISTS transactions (' \
             'timestamp TEXT, address TEXT, recipient TEXT, amount TEXT, signature TEXT, ' \
             "public_key TEXT, operation TEXT, openfield TEXT, mergedts INTEGER(4) not null default (strftime('%s','now')) )"

# Purge old txs that may be stuck
SQL_PURGE = "DELETE FROM transactions WHERE timestamp <= strftime('%s', 'now', '-2 hour')"

# Delete all transactions
SQL_CLEAR = 'DELETE FROM transactions'

# Check for presence of a given tx signature
SQL_SIG_CHECK = 'SELECT timestamp FROM transactions WHERE substr(signature,1,4) = substr(?1,1,4) and signature = ?1'
SQL_SIG_CHECK_OLD = 'SELECT timestamp FROM transactions WHERE signature = ?1'

# delete a single tx
SQL_DELETE_TX = 'DELETE FROM transactions WHERE substr(signature,1,4) = substr(?1,1,4) and signature = ?1'
SQL_DELETE_TX_OLD = 'DELETE FROM transactions WHERE signature = ?1'

# Selects all tx from mempool - list fields so we don't send mergedts and keep compatibility
SQL_SELECT_ALL_TXS = 'SELECT timestamp, address, recipient, amount, signature, public_key, operation, openfield FROM transactions'

# Selects all tx from mempool - list fields so we don't send mergedts and keep compatibility
SQL_SELECT_ALL_VALID_TXS = "SELECT timestamp, address, recipient, amount, signature, public_key, operation, openfield FROM transactions WHERE timestamp > strftime('%s', 'now', '-2 hour')"

# Counts distinct senders from mempool
SQL_COUNT_DISTINCT_SENDERS = 'SELECT COUNT(DISTINCT(address)) FROM transactions'

# Counts distinct recipients from mempool
SQL_COUNT_DISTINCT_RECIPIENTS = 'SELECT COUNT(DISTINCT(recipient)) FROM transactions'

# A single requets for status info
SQL_STATUS = 'SELECT COUNT(*) AS nb, SUM(LENGTH(openfield)) AS len, COUNT(DISTINCT(address)) as senders, COUNT(DISTINCT(recipient)) as recipients FROM transactions'

# Select Tx to be sent to a peer
SQL_SELECT_TX_TO_SEND = 'SELECT * FROM transactions ORDER BY amount DESC'

# Select Tx to be sent to a peer since the given ts - what counts is the merged time, not the tx time.
SQL_SELECT_TX_TO_SEND_SINCE = 'SELECT * FROM transactions where mergedts > ? ORDER BY amount DESC'

SQL_MEMPOOL_GET = "SELECT amount, openfield, operation FROM transactions WHERE address = ? and timestamp > strftime('%s', 'now', '-2 hour')"


def sql_trace_callback(log, id, statement):
    line = f'SQL[{id}] {statement}'
    log.warning(line)


class Mempool:
    """The mempool manager. Thread safe"""
    def __init__(self, app_log, config=None, db_lock=None, testnet=False, trace_db_calls=True):
        try:
            self.app_log = app_log
            self.config = config
            self.db_lock = db_lock
            self.ram = self.config.mempool_ram
            if self.config.version == 'regnet':
                self.app_log.warning('Regtest mode, ram mempool')
                self.ram = True

            self.lock = threading.Lock()
            self.peers_lock = threading.Lock()
            # ip: last time sent
            self.peers_sent = dict()
            self.db = None
            self.cursor = None
            self.trace_db_calls = trace_db_calls

            self.testnet = testnet

            if self.testnet:
                app_log.warning('Starting mempool in testnet mode')
                self.mempool_path = 'mempool_testnet.db'
                self.mempool_ram_file = 'file:mempool_testnet?mode=memory&cache=shared'
            else:
                self.mempool_ram_file = 'file:mempool?mode=memory&cache=shared'
                self.mempool_path = self.config.mempool_path  #default

            self.check()

        except Exception as e:
            self.app_log.error('Error creating mempool: {}'.format(e))
            raise

    def mp_get(self, balance_address):
        """
        base mempool
        :return:
        """
        return self.fetchall(SQL_MEMPOOL_GET, (balance_address, ))

    def check(self):
        """
        Checks if mempool exists, create if not.
        :return:
        """
        self.app_log.debug('Mempool Check')
        with self.lock:
            if self.ram:
                self.db = sqlite3.connect(self.mempool_ram_file, uri=True, timeout=1, isolation_level=None, check_same_thread=False)
                if self.trace_db_calls:
                    self.db.set_trace_callback(functools.partial(sql_trace_callback, self.app_log, 'MEMPOOL-RAM'))
                self.db.execute('PRAGMA journal_mode = WAL;')
                self.db.execute('PRAGMA page_size = 4096;')
                self.db.text_factory = str
                self.cursor = self.db.cursor()
                self.cursor.execute(SQL_CREATE)
                self.db.commit()
                self.app_log.warning('Status: In memory mempool file created')
            else:
                self.db = sqlite3.connect(self.mempool_path, timeout=1, check_same_thread=False)
                if self.trace_db_calls:
                    self.db.set_trace_callback(functools.partial(sql_trace_callback, self.app_log, 'MEMPOOL'))
                self.db.text_factory = str
                self.cursor = self.db.cursor()

                # check if mempool needs recreating
                self.cursor.execute("PRAGMA table_info('transactions')")
                res = self.cursor.fetchall()
                # print(res)
                if len(res) != 9:
                    self.db.close()
                    os.remove(self.mempool_path)
                    self.db = sqlite3.connect(self.mempool_path, timeout=1, check_same_thread=False)
                    if self.trace_db_calls:
                        self.db.set_trace_callback(functools.partial(sql_trace_callback, self.app_log, 'MEMPOOL'))
                    self.db.text_factory = str
                    self.cursor = self.db.cursor()
                    self.execute(SQL_CREATE)
                    self.commit()
                    self.app_log.warning('Status: Recreated mempool file')

    def execute(self, sql, param=None, cursor=None):
        """
        Safely execute the request
        :param sql:
        :param param:
        :param cursor: optional. will use the locked shared cursor if None
        :return:
        """
        # TODO: add a try count and die if we lock
        while True:
            try:
                if not cursor:
                    cursor = self.cursor
                if param:
                    cursor.execute(sql, param)
                else:
                    cursor.execute(sql)
                break
            except Exception as e:
                self.app_log.warning('Database query: {} {}'.format(cursor, sql))
                self.app_log.warning('Database retry reason: {}'.format(e))
                time.sleep(0.1)

    def commit(self):
        """
        Safe commit
        :return:
        """
        # no lock on execute and commit. locks are on full atomic operations only
        while True:
            try:
                self.db.commit()
                break
            except Exception as e:
                self.app_log.warning('Database retry reason: {}'.format(e))
                time.sleep(0.1)

    def fetchone(self, sql, param=None, write=False):
        """
        Fetchs one and Returns data
        :param sql:
        :param param:
        :param write: if the requests involves write, set to True to request a Lock
        :return:
        """
        if write:
            with self.lock:
                self.execute(sql, param)
                return self.cursor.fetchone()
        else:
            cursor = self.db.cursor()
            self.execute(sql, param, cursor)
            return cursor.fetchone()

    def fetchall(self, sql, param=None, write=False):
        """
        Fetchs all and Returns data
        :param sql:
        :param param:
        :param write: if the requests involves write, set to True to request a Lock
        :return:
        """
        if write:
            with self.lock:
                self.execute(sql, param)
                return self.cursor.fetchall()
        else:
            cursor = self.db.cursor()
            self.execute(sql, param, cursor)
            return cursor.fetchall()

    def vacuum(self):
        """
        Maintenance
        :return:
        """
        with self.lock:
            self.execute('VACUUM')

    def close(self):
        if self.db:
            self.db.close()

    def purge(self):
        """
        Purge old txs
        :return:
        """
        with self.lock:
            self.app_log.warning('Purging mempool')
            try:
                self.execute(SQL_PURGE)
                self.commit()
            except Exception as e:
                self.app_log.error('Error {} on mempool purge'.format(e))

    def clear(self):
        """
        Empty mempool
        :return:
        """
        with self.lock:
            self.execute(SQL_CLEAR)
            self.commit()

    def delete_transaction(self, signature):
        """
        Delete a single tx by its id
        :return:
        """
        with self.lock:
            if self.config.old_sqlite:
                self.execute(SQL_DELETE_TX_OLD, (signature, ))
            else:
                self.execute(SQL_DELETE_TX, (signature, ))
            self.commit()

    def sig_check(self, signature):
        """
        Returns presence of the sig in the mempool
        :param signature:
        :return: boolean
        """
        if self.config.old_sqlite:
            return bool(self.fetchone(SQL_SIG_CHECK_OLD, (signature, )))
        else:
            return bool(self.fetchone(SQL_SIG_CHECK, (signature, )))

    def status(self):
        """
        Stats on the current mempool
        :return: tuple(tx#, openfield len, distinct sender#, distinct recipients#
        """
        try:
            limit = time.time()
            frozen = [peer for peer in self.peers_sent if self.peers_sent[peer] > limit]
            # self.app_log.warning('Status: MEMPOOL Frozen = {}'.format(', '.join(frozen)))
            # print(limit, self.peers_sent, frozen)
            # Cleanup old nodes not synced since 15 min
            limit = limit - 15*60
            with self.peers_lock:
                self.peers_sent = {peer: self.peers_sent[peer] for peer in self.peers_sent if self.peers_sent[peer] > limit}
            # self.app_log.warning('Status: MEMPOOL Live = {}'.format(', '.join(set(self.peers_sent.keys()) - set(frozen))))
            status = self.fetchall(SQL_STATUS)
            count, open_len, senders, recipients = status[0]
            self.app_log.debug('Status: MEMPOOL {} Txs from {} senders to {} distinct recipients. Openfield len {}'.format(count, senders, recipients, open_len))
            return status[0]
        except:
            return 0

    def size(self):
        """
        Curent size of the mempool in Mo
        :return:
        """
        try:
            mempool_txs = self.fetchall(SQL_SELECT_ALL_VALID_TXS)
            mempool_size = sys.getsizeof(str(mempool_txs))/1000000.0
            return mempool_size
        except:
            return 0

    def sent(self, peer_ip):
        """
        record time of last mempool send to this peer
        :param peer_ip:
        :return:
        """
        # TODO: have a purge
        when = time.time()
        if peer_ip in self.peers_sent:
            # can be frozen, no need to lock and update, time is already in the future.
            if self.peers_sent[peer_ip] > when:
                return
        with self.peers_lock:
            self.peers_sent[peer_ip] = when

    def sendable(self, peer_ip):
        """
        Tells is the mempool is sendable to a given peers
        (ie, we sent it more than 30 sec ago)
        :param peer_ip:
        :return:
        """
        if peer_ip not in self.peers_sent:
            # New peer
            return True
        sendable = self.peers_sent[peer_ip] < time.time() - 30
        # Temp
        if not sendable:
            pass
            # self.app_log.warning("Mempool not sendable for {} yet.".format(peer_ip))
        return sendable

    def tx_to_send(self, peer_ip, peer_txs=None):
        """
        Selects the Tx to be sent to a given peer
        :param peer_ip:
        :return:
        """
        if DEBUG_DO_NOT_SEND_TX:
            all = self.fetchall(SQL_SELECT_TX_TO_SEND)
            tx_count = len(all)
            tx_list = [tx[1] + ' ' + tx[2] + ' : ' + str(tx[3]) for tx in all]
            # print("I have {} txs for {} but won't send: {}".format(tx_count, peer_ip, "\n".join(tx_list)))
            # print("I have {} txs for {} but won't send".format(tx_count, peer_ip))
            return []
        # Get our raw txs
        if peer_ip not in self.peers_sent:
            # new peer, never seen, send all
            raw = self.fetchall(SQL_SELECT_TX_TO_SEND)
        else:
            # add some margin to account for tx in the future, 5 sec ?
            last_sent = self.peers_sent[peer_ip] - 5
            raw = self.fetchall(SQL_SELECT_TX_TO_SEND_SINCE, (last_sent, ))
        # Now filter out the tx we got from the peer
        if peer_txs:
            peers_sig = [tx[4] for tx in peer_txs]
            # TEMP
            # print("raw for", peer_ip, len(raw))
            # print("peers_sig", peer_ip, len(peers_sig))

            filtered = [tx for tx in raw if tx[4] not in peers_sig]
            # TEMP
            # print("filtered", peer_ip, len(filtered))
            return filtered
        else:
            return raw

    def space_left_for_tx(self, transaction, mempool_size):
        """
        Tells if we should let a specific tx in, depending on space left and its characteristics.
        :param transaction:
        :param mempool_size:
        :param size_bypass:
        :return:
        """
        # Allow whatever the tx is
        if mempool_size < 0.3:
            return True
        # Low priority tx, token or openfield data
        if mempool_size < 0.4:
            if len(str(transaction[7])) > 200:
                # Openfield > 200
                return True
            if 'token:' == transaction[6][:6]:
                return True
        # Medium prio: 5 BIS or more
        if mempool_size < 0.5:
            if float(transaction[3]) > 5:
                return True
        # High prio: allowed by config
        if mempool_size < 0.6:
            if transaction[1] in self.config.mempool_allowed:
                return True
        # Sorry, no space left for this tx type.
        return False

    def merge(self, data: list, peer_ip: str, c, size_bypass: bool = False, wait: bool = False, revert: bool = False) -> list:
        """
        Checks and merge the tx list in our mempool
        :param data:
        :param peer_ip:
        :param c:
        :param size_bypass: if True, will merge whatever the mempool size is
        :param wait: if True, will wait until the main db_lock is free. if False, will just drop.
        :param revert: if True, we are reverting tx from digest_block, so main lock is on. Don't bother, process without lock.
        :return:
        """
        global REFUSE_OLDER_THAN
        # Easy cases of empty or invalid data
        if not data:
            return ['Mempool from {} was empty'.format(peer_ip)]
        mempool_result = []
        if data == '*':
            raise ValueError('Connection lost')
        try:
            if self.peers_sent[peer_ip] > time.time() and peer_ip != '127.0.0.1':
                self.app_log.warning('Mempool ignoring merge from frozen {}'.format(peer_ip))
                mempool_result.append('Mempool ignoring merge from frozen {}'.format(peer_ip))
                return mempool_result
        except:
            # unknown peer
            pass
        if not essentials.is_sequence(data):
            if peer_ip != '127.0.0.1':
                with self.peers_lock:
                    self.peers_sent[peer_ip] = time.time() + 10*60
                self.app_log.warning('Freezing mempool from {} for 10 min - Bad TX format'.format(peer_ip))
            mempool_result.append('Bad TX Format')
            return mempool_result

        if not revert:
            while self.db_lock.locked():
                # prevent transactions which are just being digested from being added to mempool
                if not wait:
                    # not reverting, but not waiting, bye
                    # By default, we don't wait.
                    mempool_result.append('Locked ledger, dropping txs')
                    return mempool_result
                self.app_log.warning('Waiting for block digestion to finish before merging mempool')
                time.sleep(1)
        # if reverting, don't bother with main lock, go on.
        # Let's really dig
        mempool_result.append('Mempool merging started from {}'.format(peer_ip))
        # Single time reference here for the whole merge.
        time_now = time.time()
        # calculate current mempool size before adding txs
        mempool_size = self.size()

        # TODO: we check main ledger db is not locked before beginning, but we don't lock? ok, see comment in node.py. since it's called from a lock, it would deadlock.
        # merge mempool
        # while self.lock.locked():
        #    time.sleep(1)
        with self.lock:
            try:
                block_list = data
                if not isinstance(block_list[0], list):  # convert to list of lists if only one tx and not handled
                    block_list = [block_list]

                for transaction in block_list:
                    if size_bypass or self.space_left_for_tx(transaction, mempool_size):
                        # all transactions in the mempool need to be cycled to check for special cases,
                        # therefore no while/break loop here
                        try:
                            mempool_timestamp = '%.2f' % (quantize_two(transaction[0]))
                            mempool_timestamp_float = float(transaction[0])  # limit Decimal where not needed
                        except Exception as e:
                            mempool_result.append('Mempool: Invalid timestamp {}'.format(transaction[0]))
                        if not essentials.address_validate(transaction[1]):
                            mempool_result.append('Mempool: Invalid address {}'.format(transaction[1]))
                            continue
                        # We could now ignore the truncates here, I left them for explicit reminder of the various fields max lengths.
                        mempool_address = str(transaction[1])[:56]
                        if not essentials.address_validate(transaction[2]):
                            mempool_result.append('Mempool: Invalid recipient {}'.format(transaction[2]))
                            continue
                        mempool_recipient = str(transaction[2])[:56]
                        try:
                            mempool_amount = '%.8f' % (quantize_eight(transaction[3]))  # convert scientific notation
                            mempool_amount_float = float(transaction[3])
                        except Exception as e:
                            mempool_result.append('Mempool: Invalid amount {}'.format(transaction[3]))
                            continue
                        if len(transaction[4]) > 684:
                            mempool_result.append('Mempool: Invalid signature len{}'.format(len(transaction[4])))
                            continue
                        mempool_signature_enc = str(transaction[4])[:684]
                        if len(transaction[5]) > 1068:
                            mempool_result.append('Mempool: Invalid pubkey len{}'.format(len(transaction[5])))
                            continue
                        mempool_public_key_b64encoded = str(transaction[5])[:1068]
                        if "b'" == mempool_public_key_b64encoded[:2]:
                            # Binary content instead of str - leftover from legacy code?
                            mempool_public_key_b64encoded = transaction[5][2:1070]
                        if len(transaction[6]) > 30:
                            mempool_result.append('Mempool: Invalid operation len{}'.format(len(transaction[6])))
                            continue
                        mempool_operation = str(transaction[6])[:30]
                        if len(transaction[7]) > 100000:
                            mempool_result.append('Mempool: Invalid openfield len{}'.format(len(transaction[7])))
                            continue
                        mempool_openfield = str(transaction[7])[:100000]

                        if len(mempool_openfield) <= 4:
                            # no or short message for a mandatory message
                            if mempool_recipient in self.config.mandatory_message.keys():
                                mempool_result.append('Mempool: Missing message - {}'.format(self.config.mandatory_message[mempool_recipient]))
                                continue

                        # Begin with the easy tests that do not require cpu or disk access
                        if mempool_amount_float < 0:
                            mempool_result.append('Mempool: Negative balance spend attempt')
                            continue
                        if mempool_timestamp_float > time_now:
                            mempool_result.append('Mempool: Future transaction rejected {}s'.format(mempool_timestamp_float - time_now))
                            continue
                        if mempool_timestamp_float < time_now - REFUSE_OLDER_THAN:
                            # don't accept old txs, mempool needs to be harsher than ledger
                            mempool_result.append('Mempool: Too old a transaction')
                            continue

                        # Then more cpu heavy tests
                        buffer = str((mempool_timestamp, mempool_address, mempool_recipient, mempool_amount, mempool_operation, mempool_openfield)).encode('utf-8')

                        #  Will raise if error
                        try:
                            SignerFactory.verify_bis_signature(mempool_signature_enc, mempool_public_key_b64encoded, buffer, mempool_address)
                        except Exception as e:
                            mempool_result.append(f'Mempool: Signature did not match for address ({e})')
                            continue

                        # Only now, process the tests requiring db access
                        mempool_in = self.sig_check(mempool_signature_enc)

                        # Temp: get last block for HF reason
                        essentials.execute_param_c(c, 'SELECT block_height FROM transactions WHERE 1 ORDER by block_height DESC limit ?', (1, ), self.app_log)
                        last_block = c.fetchone()[0]
                        # reject transactions which are already in the ledger
                        # TODO: not clean, will need to have ledger as a module too.
                        # TODO: need better txid index, this is very sloooooooow
                        if self.config.old_sqlite:
                            essentials.execute_param_c(c, 'SELECT timestamp FROM transactions WHERE signature = ?1', (mempool_signature_enc, ), self.app_log)
                        else:
                            essentials.execute_param_c(c, 'SELECT timestamp FROM transactions WHERE substr(signature,1,4) = substr(?1,1,4) AND signature = ?1', (mempool_signature_enc, ), self.app_log)
                        ledger_in = bool(c.fetchone())
                        # remove from mempool if it's in both ledger and mempool already
                        if mempool_in and ledger_in:
                            try:
                                # Do not lock, we already have the lock for the whole merge.
                                if self.config.old_sqlite:
                                    self.execute(SQL_DELETE_TX_OLD, (mempool_signature_enc, ))
                                else:
                                    self.execute(SQL_DELETE_TX, (mempool_signature_enc, ))
                                self.commit()
                                mempool_result.append('Mempool: Transaction deleted from our mempool')
                            except:  # experimental try and except
                                mempool_result.append('Mempool: Transaction was not present in the pool anymore')
                            continue
                        if ledger_in:
                            mempool_result.append('That transaction is already in our ledger')
                            # Can be a syncing node. Do not request mempool from this peer until FREEZE_MIN min
                            # ledger_in is the ts of the tx in ledger. if it's recent, maybe the peer is just one block late.
                            # give him 15 minute margin.
                            if (peer_ip != '127.0.0.1') and (ledger_in < time_now - 60*15):
                                with self.peers_lock:
                                    self.peers_sent[peer_ip] = time.time() + FREEZE_MIN*60
                                self.app_log.warning('Freezing mempool from {} for {} min.'.format(peer_ip, FREEZE_MIN))
                            # Here, we point blank stop processing the batch from this host since it's outdated.
                            # Update: Do not, since it blocks further valid tx - case has been found in real use.
                            # return mempool_result
                            continue
                        # Already there, just ignore then
                        if mempool_in:
                            mempool_result.append('That transaction is already in our mempool')
                            continue

                        # Here we covered the basics, the current tx is conform and signed. Now let's check balance.

                        # verify balance
                        mempool_result.append('Mempool: Received address: {}'.format(mempool_address))
                        # include mempool fees - excluding the old ones.
                        result = self.fetchall(SQL_MEMPOOL_GET, (mempool_address, ))
                        debit_mempool = DECIMAL0
                        if result:
                            for x in result:
                                debit_tx = quantize_eight(x[0])
                                fee = essentials.fee_calculate(x[1], x[2], last_block)  # fee_calculate sends back a Decimal 8
                                debit_mempool += debit_tx + fee

                        credit = DECIMAL0
                        rewards = DECIMAL0
                        for entry in essentials.execute_param_c(c, 'SELECT amount, reward FROM transactions WHERE recipient = ?', (mempool_address, ), self.app_log):
                            credit += quantize_eight(entry[0])
                            rewards += quantize_eight(entry[1])

                        debit_ledger = DECIMAL0
                        fees = DECIMAL0
                        for entry in essentials.execute_param_c(c, 'SELECT amount, fee FROM transactions WHERE address = ?', (mempool_address, ), self.app_log):
                            debit_ledger += quantize_eight(entry[0])
                            fees += quantize_eight(entry[1])

                        debit = debit_ledger + debit_mempool

                        # both are Decimals
                        balance = credit - debit - fees + rewards - quantize_eight(mempool_amount)
                        balance_pre = credit - debit_ledger - fees + rewards

                        fee = essentials.fee_calculate(mempool_openfield, mempool_operation, last_block)

                        # print("Balance", balance, fee)

                        if quantize_eight(mempool_amount) > balance_pre:  # mempool_amount is a 0.8f string for some reason
                            # mp amount is already included in "balance" var! also, that tx might already be in the mempool
                            mempool_result.append('Mempool: Sending more than owned')
                            continue
                        if balance - fee < 0:
                            mempool_result.append('Mempool: Cannot afford to pay fees')
                            continue

                        # Pfew! we can finally insert into mempool - all is str, type converted and enforced above
                        self.execute(
                            'INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?)',
                            (mempool_timestamp, mempool_address, mempool_recipient, mempool_amount, mempool_signature_enc, mempool_public_key_b64encoded, mempool_operation, mempool_openfield, int(time_now)),
                        )
                        mempool_result.append('Mempool updated with a received transaction from {}'.format(peer_ip))
                        mempool_result.append('Success')  # WARNING: Do not change string or case ever!
                        self.commit()  # Save (commit) the changes to mempool db

                        mempool_size += sys.getsizeof(str(transaction))/1000000.0
                    else:
                        mempool_result.append('Local mempool is already full for this tx type, skipping merging')
                        # self.app_log.warning("Local mempool is already full for this tx type, skipping merging")
                # TEMP
                # print("Mempool insert", mempool_result)
                return mempool_result
                # TODO: Here maybe commit() on c to release the write lock?
            except Exception as e:
                self.app_log.warning('Mempool: Error processing: {} {}'.format(data, e))
                if self.config.debug:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                    self.app_log.warning('{} {} {}'.format(exc_type, fname, exc_tb.tb_lineno))
                    mempool_result.append('Exception: {}'.format(str(e)))
                    # if left there, means debug can *not* be used in production, or exception is not sent back to the client.
                    raise
        return mempool_result



#!/usr/bin/env python
# rebuild.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (rebuild.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

if __name__ == '__main__':
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))

#------------------------------------------------------------------------------

import bitdust.logs.lg

import bitdust.raid.read
import bitdust.raid.eccmap

#------------------------------------------------------------------------------


def rebuild(backupID, blockNum, eccMap, availableSuppliers, remoteMatrix, localMatrix, localBackupsDir, threshold_control=None):
    try:
        if _Debug:
            with open('/tmp/raid.log', 'a') as logfile:
                logfile.write(u'rebuild backupID=%r blockNum=%r eccMap=%r\n' % (
                    backupID,
                    blockNum,
                    eccMap,
                ))

        customer, _, localPath = backupID.rpartition(':')
        if '$' not in customer:
            customer = 'master$' + customer
        myeccmap = bitdust.raid.eccmap.eccmap(eccMap)
        supplierCount = len(availableSuppliers)
        missingData = [0]*supplierCount
        missingParity = [0]*supplierCount
        reconstructedData = [0]*supplierCount
        reconstructedParity = [0]*supplierCount
        remoteData = list(remoteMatrix['D'])
        remoteParity = list(remoteMatrix['P'])
        localData = list(localMatrix['D'])
        localParity = list(localMatrix['P'])

        def _build_raid_file_name(supplierNumber, dataOrParity):
            return os.path.join(localBackupsDir, customer, localPath, str(blockNum) + '-' + str(supplierNumber) + '-' + dataOrParity)

        # This builds a list of missing pieces.
        # The file is missing if value in the corresponding cell
        # in the "remote" matrix (see ``p2p.backup_matrix``) is -1 or 0
        # but the supplier who must keep that file is online.
        # In other words, if supplier is online but do not have that piece - this piece is missing.
        for supplierNum in range(supplierCount):
            if availableSuppliers[supplierNum] == 0:
                continue
            if supplierNum >= len(remoteData) or supplierNum >= len(remoteParity):
                continue
            # if remote Data file not exist and supplier is online
            # we mark it as missing and will try to rebuild this file and send to him
            if remoteData[supplierNum] != 1:
                # mark file as missing
                missingData[supplierNum] = 1
            # same for Parity file
            if remoteParity[supplierNum] != 1:
                missingParity[supplierNum] = 1

        if _Debug:
            with open('/tmp/raid.log', 'a') as logfile:
                logfile.write(u'missingData=%r\n' % missingData)
                logfile.write(u'missingParity=%r\n' % missingParity)
                logfile.write(u'localData=%r\n' % localData)
                logfile.write(u'localParity=%r\n' % localParity)

        # This made an attempt to rebuild the missing pieces
        # from pieces we have on hands.
        # lg.out(14, 'block_rebuilder.AttemptRebuild %s %d BEGIN' % (self.backupID, self.blockNum))
        newData = False
        madeProgress = True
        while madeProgress:
            madeProgress = False
            # will check all data packets we have
            for supplierNum in range(supplierCount):
                dataFileName = _build_raid_file_name(supplierNum, 'Data')
                # if we do not have this item on hands - we will reconstruct it from other items
                if localData[supplierNum] == 0:
                    parityNum, parityMap = myeccmap.GetDataFixPath(localData, localParity, supplierNum)
                    if parityNum != -1:
                        rebuildFileList = []
                        rebuildFileList.append(_build_raid_file_name(parityNum, 'Parity'))
                        for supplierParity in parityMap:
                            if supplierParity != supplierNum:
                                filename = _build_raid_file_name(supplierParity, 'Data')
                                if os.path.isfile(filename):
                                    rebuildFileList.append(filename)
                        # lg.out(10, '    rebuilding file %s from %d files' % (os.path.basename(dataFileName), len(rebuildFileList)))
                        bitdust.raid.read.RebuildOne(rebuildFileList, len(rebuildFileList), dataFileName, threshold_control=threshold_control)
                    if os.path.exists(dataFileName):
                        localData[supplierNum] = 1
                        madeProgress = True
                        # lg.out(10, '        Data file %s found after rebuilding for supplier %d' % (os.path.basename(dataFileName), supplierNum))
                # now we check again if we have the data on hand after rebuild at it is missing - send it
                # but also check to not duplicate sending to this man
                # now sending is separated, see the file data_sender.py
                if localData[supplierNum] == 1 and missingData[supplierNum] == 1:  # and self.dataSent[supplierNum] == 0:
                    # lg.out(10, '            rebuilt a new Data for supplier %d' % supplierNum)
                    newData = True
                    reconstructedData[supplierNum] = 1
                    # self.outstandingFilesList.append((dataFileName, self.BuildFileName(supplierNum, 'Data'), supplierNum))
                    # self.dataSent[supplierNum] = 1
        # now with parities ...
        for supplierNum in range(supplierCount):
            parityFileName = _build_raid_file_name(supplierNum, 'Parity')
            if localParity[supplierNum] == 0:
                parityMap = myeccmap.ParityToData[supplierNum]
                HaveAllData = True
                for segment in parityMap:
                    if localData[segment] == 0:
                        HaveAllData = False
                        break
                if HaveAllData:
                    rebuildFileList = []
                    for supplierParity in parityMap:
                        filename = _build_raid_file_name(supplierParity, 'Data')  # ??? why not 'Parity'
                        if os.path.isfile(filename):
                            rebuildFileList.append(filename)
                    # lg.out(10, '    rebuilding file %s from %d files' % (os.path.basename(parityFileName), len(rebuildFileList)))
                    bitdust.raid.read.RebuildOne(rebuildFileList, len(rebuildFileList), parityFileName, threshold_control=threshold_control)
                    if os.path.exists(parityFileName):
                        # lg.out(10, '        Parity file %s found after rebuilding for supplier %d' % (os.path.basename(parityFileName), supplierNum))
                        localParity[supplierNum] = 1
            # so we have the parity on hand and it is missing - send it
            if localParity[supplierNum] == 1 and missingParity[supplierNum] == 1:  # and self.paritySent[supplierNum] == 0:
                # lg.out(10, '            rebuilt a new Parity for supplier %d' % supplierNum)
                newData = True
                reconstructedParity[supplierNum] = 1
                # self.outstandingFilesList.append((parityFileName, self.BuildFileName(supplierNum, 'Parity'), supplierNum))
                # self.paritySent[supplierNum] = 1
        # lg.out(14, 'block_rebuilder.AttemptRebuild END')

        if _Debug:
            with open('/tmp/raid.log', 'a') as logfile:
                logfile.write(u'newData=%r\n' % newData)

        return (
            newData,
            localData,
            localParity,
            reconstructedData,
            reconstructedParity,
        )

    except:
        bitdust.logs.lg.exc()
        return None



#!/usr/bin/env python
# routingtable.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from __future__ import print_function
import time
import random

from . import constants  # @UnresolvedImport
from . import kbucket  # @UnresolvedImport
from .protocol import TimeoutError  # @UnresolvedImport

_Debug = False


class RoutingTable(object):
    """
    Interface for RPC message translators/formatters.

    Classes inheriting from this should provide a suitable routing table
    for a parent Node object (i.e. the local entity in the Kademlia
    network)
    """
    def __init__(self, parentNodeID):
        """
        @param parentNodeID: The 160-bit node ID of the node to which this
                             routing table belongs
        @type parentNodeID: str
        """

    def addContact(self, contact):
        """
        Add the given contact to the correct k-bucket; if it already exists,
        its status will be updated.

        @param contact: The contact to add to this node's k-buckets
        @type contact: kademlia.contact.Contact
        """

    def distance(self, keyOne, keyTwo):
        """
        Calculate the XOR result between two string variables.

        @return: XOR result of two long variables
        @rtype: long
        """
        valKeyOne = int(keyOne, 16)
        valKeyTwo = int(keyTwo, 16)
        return valKeyOne ^ valKeyTwo

    def findCloseNodes(self, key, count, _rpcNodeID=None):
        """
        Finds a number of known nodes closest to the node/value with the
        specified key.

        @param key: the 160-bit key (i.e. the node or value ID) to search for
        @type key: str
        @param count: the amount of contacts to return
        @type count: int
        @param _rpcNodeID: Used during RPC, this is be the sender's Node ID
                           Whatever ID is passed in the parameter will get
                           excluded from the list of returned contacts.
        @type _rpcNodeID: str

        @return: A list of node contacts (C{kademlia.contact.Contact instances})
                 closest to the specified key.
                 This method will return C{k} (or C{count}, if specified)
                 contacts if at all possible; it will only return fewer if the
                 node is returning all of the contacts that it knows of.
        @rtype: list
        """

    def getContact(self, contactID):
        """
        Returns the (known) contact with the specified node ID.

        @raise ValueError: No contact with the specified contact ID is known
                           by this node
        """

    def getRefreshList(self, startIndex=0, force=False):
        """
        Finds all k-buckets that need refreshing, starting at the k-bucket with
        the specified index, and returns IDs to be searched for in order to
        refresh those k-buckets.

        @param startIndex: The index of the bucket to start refreshing at;
                           this bucket and those further away from it will
                           be refreshed. For example, when joining the
                           network, this node will set this to the index of
                           the bucket after the one containing it's closest
                           neighbour.
        @type startIndex: index
        @param force: If this is C{True}, all buckets (in the specified range)
                      will be refreshed, regardless of the time they were last
                      accessed.
        @type force: bool

        @return: A list of node ID's that the parent node should search for
                 in order to refresh the routing Table
        @rtype: list
        """

    def removeContact(self, contactID):
        """
        Remove the contact with the specified node ID from the routing table.

        @param contactID: The node ID of the contact to remove
        @type contactID: str
        """

    def touchKBucket(self, key):
        """
        Update the "last accessed" timestamp of the k-bucket which covers the
        range containing the specified key in the key/ID space.

        @param key: A key in the range of the target k-bucket
        @type key: str
        """


class TreeRoutingTable(RoutingTable):
    """
    This class implements a routing table used by a Node class.

    The Kademlia routing table is a binary tree whose leaves are k-buckets,
    where each k-bucket contains nodes with some common prefix of their IDs.
    This prefix is the k-bucket's position in the binary tree; it therefore
    covers some range of ID values, and together all of the k-buckets cover
    the entire 160-bit ID (or key) space (with no overlap).

    @note: In this implementation, nodes in the tree (the k-buckets) are
    added dynamically, as needed; this technique is described in the 13-page
    version of the Kademlia paper, in section 2.4. It does, however, use the
    C{PING} RPC-based k-bucket eviction algorithm described in section 2.2 of
    that paper.
    """
    def __init__(self, parentNodeID, **kwargs):
        """
        @param parentNodeID: The 160-bit node ID of the node to which this
                             routing table belongs
        @type parentNodeID: str
        """
        # Create the initial (single) k-bucket covering the range of the entire 160-bit ID space
        self._buckets = [
            kbucket.KBucket(rangeMin=0, rangeMax=2**160),
        ]
        self._parentNodeID = parentNodeID
        self._layerID = kwargs.get('layerID', 0)

    def __repr__(self, *args, **kwargs):
        return str(self)

    def __str__(self):
        return '<RTable(%d) %d buckets for %r>' % (self._layerID, len(self._buckets), self._parentNodeID)

    def totalContacts(self):
        counter = 0
        for bucket in self._buckets:
            counter += len(bucket._contacts)
        return counter

    def addContact(self, contact):
        """
        Add the given contact to the correct k-bucket; if it already exists,
        its status will be updated.

        @param contact: The contact to add to this node's k-buckets
        @type contact: kademlia.contact.Contact
        """
        if contact.id == self._parentNodeID:
            return

        bucketIndex = self._kbucketIndex(contact.id)
        try:
            self._buckets[bucketIndex].addContact(contact)
            if _Debug:
                print('[DHT RTABLE] layerID=%d   added %r to bucket %d' % (self._layerID, contact, bucketIndex))
        except kbucket.BucketFull:
            # The bucket is full; see if it can be split (by checking if its range includes the host node's id)
            if self._buckets[bucketIndex].keyInRange(self._parentNodeID):
                self._splitBucket(bucketIndex)
                # Retry the insertion attempt
                self.addContact(contact)
            else:
                # We can't split the k-bucket
                # NOTE:
                # In section 2.4 of the 13-page version of the Kademlia paper, it is specified that
                # in this case, the new contact should simply be dropped. However, in section 2.2,
                # it states that the head contact in the k-bucket (i.e. the least-recently seen node)
                # should be pinged - if it does not reply, it should be dropped, and the new contact
                # added to the tail of the k-bucket. This implementation follows section 2.2 regarding
                # this point.
                headContact = self._buckets[bucketIndex]._contacts[0]

                def replaceContact(failure):
                    """
                    Callback for the deferred PING RPC to see if the head node
                    in the k-bucket is still responding.

                    @type failure: twisted.python.failure.Failure
                    """
                    failure.trap(TimeoutError)
                    # Remove the old contact...
                    deadContactID = failure.getErrorMessage()
                    if _Debug:
                        print('[DHT RTABLE] layerID=%d   replacing dead contact %r' % (
                            self._layerID,
                            deadContactID,
                        ), )
                    try:
                        self._buckets[bucketIndex].removeContact(deadContactID)
                    except ValueError:
                        # The contact has already been removed (probably due to a timeout)
                        pass
                    # ...and add the new one at the tail of the bucket
                    self.addContact(contact)

                # Ping the least-recently seen contact in this k-bucket
                headContact = self._buckets[bucketIndex]._contacts[0]
                df = headContact.ping()
                # If there's an error (i.e. timeout), remove the head contact, and append the new one
                df.addErrback(replaceContact)

    def findCloseNodes(self, key, count, _rpcNodeID=None):
        """
        Finds a number of known nodes closest to the node/value with the
        specified key.

        @param key: the 160-bit key (i.e. the node or value ID) to search for
        @type key: str
        @param count: the amount of contacts to return
        @type count: int
        @param _rpcNodeID: Used during RPC, this is be the sender's Node ID
                           Whatever ID is passed in the paramater will get
                           excluded from the list of returned contacts.
        @type _rpcNodeID: str

        @return: A list of node contacts (C{kademlia.contact.Contact instances})
                 closest to the specified key.
                 This method will return C{k} (or C{count}, if specified)
                 contacts if at all possible; it will only return fewer if the
                 node is returning all of the contacts that it knows of.
        @rtype: list
        """
        # if key == self.id:
        #    bucketIndex = 0 #TODO: maybe not allow this to continue?
        # else:
        bucketIndex = self._kbucketIndex(key)

        if _Debug:
            print('[DHT RTABLE] layerID=%d   findCloseNodes %r   _rpcNodeID=%r   bucketIndex=%d buckets=%d' % (
                self._layerID,
                key,
                _rpcNodeID if _rpcNodeID else None,
                bucketIndex,
                len(self._buckets),
            ))

        closestNodes = self._buckets[bucketIndex].getContacts(constants.k, _rpcNodeID)
        # This method must return k contacts (even if we have the node with the specified key as node ID),
        # unless there is less than k remote nodes in the routing table
        i = 1
        canGoLower = bucketIndex - i >= 0
        canGoHigher = bucketIndex + i < len(self._buckets)
        # Fill up the node list to k nodes, starting with the closest neighbouring nodes known
        while len(closestNodes) < constants.k and (canGoLower or canGoHigher):
            if _Debug:
                print('[DHT RTABLE] layerID=%d  closestNodes=%r' % (
                    self._layerID,
                    closestNodes,
                ))
            # TODO: this may need to be optimized
            more_contacts = []
            if canGoLower:
                more_contacts = self._buckets[bucketIndex - i].getContacts(constants.k - len(closestNodes), _rpcNodeID)
                closestNodes.extend(more_contacts)
                canGoLower = bucketIndex - (i + 1) >= 0
            if canGoHigher:
                more_contacts = self._buckets[bucketIndex + i].getContacts(constants.k - len(closestNodes), _rpcNodeID)
                closestNodes.extend(more_contacts)
                canGoHigher = bucketIndex + (i + 1) < len(self._buckets)
            i += 1
            if _Debug:
                print('[DHT RTABLE] layerID=%d   canGoLower=%s canGoHigher=%s more_contacts=%r' % (
                    self._layerID,
                    canGoLower,
                    canGoHigher,
                    more_contacts,
                ))
        if _Debug:
            print('[DHT RTABLE] layerID=%d   result=%r' % (
                self._layerID,
                closestNodes,
            ))
        return closestNodes

    def getContact(self, contactID):
        """
        Returns the (known) contact with the specified node ID.

        @raise ValueError: No contact with the specified contact ID is known
                           by this node
        """
        bucketIndex = self._kbucketIndex(contactID)
        try:
            contact = self._buckets[bucketIndex].getContact(contactID)
        except ValueError:
            raise
        else:
            return contact

    def getRefreshList(self, startIndex=0, force=False):
        """
        Finds all k-buckets that need refreshing, starting at the k-bucket with
        the specified index, and returns IDs to be searched for in order to
        refresh those k-buckets.

        @param startIndex: The index of the bucket to start refreshing at;
                           this bucket and those further away from it will
                           be refreshed. For example, when joining the
                           network, this node will set this to the index of
                           the bucket after the one containing it's closest
                           neighbour.
        @type startIndex: index
        @param force: If this is C{True}, all buckets (in the specified range)
                      will be refreshed, regardless of the time they were last
                      accessed.
        @type force: bool

        @return: A list of node ID's that the parent node should search for
                 in order to refresh the routing Table
        @rtype: list
        """
        bucketIndex = startIndex
        refreshIDs = []
        for bucket in self._buckets[startIndex:]:
            if force or (int(time.time()) - bucket.lastAccessed >= constants.refreshTimeout):
                searchID = self._randomIDInBucketRange(bucketIndex)
                refreshIDs.append(searchID)
            bucketIndex += 1
        return refreshIDs

    def removeContact(self, contactID):
        """
        Remove the contact with the specified node ID from the routing table.

        @param contactID: The node ID of the contact to remove
        @type contactID: str
        """
        bucketIndex = self._kbucketIndex(contactID)
        try:
            self._buckets[bucketIndex].removeContact(contactID)
            if _Debug:
                print('[DHT RTABLE] layerID=%d   removeContact(%r) from bucket %d' % (self._layerID, contactID, bucketIndex))
        except ValueError:
            if _Debug:
                print('[DHT RTABLE] layerID=%d   removeContact(%r): Contact not in routing table bucketIndex=%r' % (self._layerID, contactID, bucketIndex))
            return

    def touchKBucket(self, key):
        """
        Update the "last accessed" timestamp of the k-bucket which covers the
        range containing the specified key in the key/ID space.

        @param key: A key in the range of the target k-bucket
        @type key: str
        """
        bucketIndex = self._kbucketIndex(key)
        self._buckets[bucketIndex].lastAccessed = int(time.time())

    def _kbucketIndex(self, key):
        """
        Calculate the index of the k-bucket which is responsible for the
        specified key (or ID)

        @param key: The key for which to find the appropriate k-bucket index
        @type key: str

        @return: The index of the k-bucket responsible for the specified key
        @rtype: int
        """
        valKey = int(key, 16)
        i = 0
        for bucket in self._buckets:
            if bucket.keyInRange(valKey):
                # if _Debug:
                #     print('[DHT RTABLE] _kbucketIndex  %r  %r  returning %r' % (key, valKey, i, ))
                return i
            else:
                i += 1
        # if _Debug:
        #     print('[DHT RTABLE] _kbucketIndex  %r  %r  finishing with %r' % (key, valKey, i, ))
        return i

    def _randomIDInBucketRange(self, bucketIndex):
        """
        Returns a random ID in the specified k-bucket's range.

        @param bucketIndex: The index of the k-bucket to use
        @type bucketIndex: int
        """
        idValue = random.randrange(self._buckets[bucketIndex].rangeMin, self._buckets[bucketIndex].rangeMax)
        randomID = hex(idValue)[2:]
        if randomID[-1] == 'L':
            randomID = randomID[:-1]
        if len(randomID) % 2 != 0:
            randomID = '0' + randomID
        randomID = (40 - len(randomID))*'0' + randomID
        return randomID

    def _splitBucket(self, oldBucketIndex):
        """
        Splits the specified k-bucket into two new buckets which together cover
        the same range in the key/ID space.

        @param oldBucketIndex: The index of k-bucket to split (in this table's
                               list of k-buckets)
        @type oldBucketIndex: int
        """
        # Resize the range of the current (old) k-bucket
        oldBucket = self._buckets[oldBucketIndex]
        oldCount = len(oldBucket)
        splitPoint = oldBucket.rangeMax - (oldBucket.rangeMax - oldBucket.rangeMin)/2
        # Create a new k-bucket to cover the range split off from the old bucket
        newBucket = kbucket.KBucket(splitPoint, oldBucket.rangeMax)
        oldBucket.rangeMax = splitPoint
        # Now, add the new bucket into the routing table tree
        self._buckets.insert(oldBucketIndex + 1, newBucket)
        # Finally, copy all nodes that belong to the new k-bucket into it...
        for contact in oldBucket._contacts:
            if newBucket.keyInRange(contact.id):
                newBucket.addContact(contact)
        # ...and remove them from the old bucket
        for contact in newBucket._contacts:
            oldBucket.removeContact(contact)
        if _Debug:
            print('[DHT RTABLE] layerID=%d   split bucket %d,    old: %d     new: %d / %d' % (
                self._layerID,
                oldBucketIndex,
                oldCount,
                len(oldBucket),
                len(newBucket),
            ))


class OptimizedTreeRoutingTable(TreeRoutingTable):
    """
    A version of the "tree"-type routing table specified by Kademlia, along
    with contact accounting optimizations specified in section 4.1 of of the
    13-page version of the Kademlia paper.
    """
    def __init__(self, parentNodeID):
        TreeRoutingTable.__init__(self, parentNodeID)
        # Cache containing nodes eligible to replace stale k-bucket entries
        self._replacementCache = {}

    def addContact(self, contact):
        """
        Add the given contact to the correct k-bucket; if it already exists,
        its status will be updated.

        @param contact: The contact to add to this node's k-buckets
        @type contact: kademlia.contact.Contact
        """

        if contact.id == self._parentNodeID:
            return

        # Initialize/reset the "successively failed RPC" counter
        contact.failedRPCs = 0

        bucketIndex = self._kbucketIndex(contact.id)
        if _Debug:
            print('[DHT RTABLE]   addContact %r at %r' % (contact.id, bucketIndex))
        try:
            self._buckets[bucketIndex].addContact(contact)
        except kbucket.BucketFull:
            if _Debug:
                print('[DHT RTABLE]    BucketFull!')
            # The bucket is full; see if it can be split (by checking if its range includes the host node's id)
            if self._buckets[bucketIndex].keyInRange(self._parentNodeID):
                self._splitBucket(bucketIndex)
                # Retry the insertion attempt
                self.addContact(contact)
            else:
                # We can't split the k-bucket
                # NOTE: This implementation follows section 4.1 of the 13 page version
                # of the Kademlia paper (optimized contact accounting without PINGs
                #- results in much less network traffic, at the expense of some memory)

                # Put the new contact in our replacement cache for the corresponding k-bucket (or update it's position if it exists already)
                if bucketIndex not in self._replacementCache:
                    self._replacementCache[bucketIndex] = []
                if contact in self._replacementCache[bucketIndex]:
                    self._replacementCache[bucketIndex].remove(contact)
                # TODO: Using k to limit the size of the contact replacement cache - maybe define a seperate value for this in constants.py?
                elif len(self._replacementCache) >= constants.k:
                    self._replacementCache.pop(0)
                self._replacementCache[bucketIndex].append(contact)

    def removeContact(self, contactID):
        """
        Remove the contact with the specified node ID from the routing table.

        @param contactID: The node ID of the contact to remove
        @type contactID: str
        """
        bucketIndex = self._kbucketIndex(contactID)
        try:
            contact = self._buckets[bucketIndex].getContact(contactID)
        except ValueError:
            # print 'removeContact(): Contact not in routing table'
            return
        contact.failedRPCs += 1
        if contact.failedRPCs >= 5:
            self._buckets[bucketIndex].removeContact(contactID)
            # Replace this stale contact with one from our replacemnent cache, if we have any
            if bucketIndex in self._replacementCache:
                if len(self._replacementCache[bucketIndex]) > 0:
                    self._buckets[bucketIndex].addContact(self._replacementCache[bucketIndex].pop())



#!/usr/bin/python
# dirsize.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (dirsize.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: dirsize.

Here is a tool to calculate the whole folder size. You start a thread
and it will do the job and than remember that size. Now you have a fast
way to get the folder size, you can ask to scan same folder again.
"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in dirsize.py')

from twisted.internet import threads

from bitdust.logs import lg

from bitdust.lib import diskspace

from bitdust.system import bpio

#------------------------------------------------------------------------------

_Jobs = {}
_Dirs = {}

#------------------------------------------------------------------------------


def ask(dirpath, callback=None, arg=None):
    """
    Start a thread to scan all sub folders and calculate total size of given
    directory.

    :param callback: set a callback function to get the folder size in your code
    :param arg: set some argument to put in the callback, so you can mark that result
    """
    global _Jobs
    global _Dirs
    if _Debug:
        lg.out(_DebugLevel, 'dirsize.ask %s' % dirpath)
    if dirpath in _Jobs:
        return 'counting size'
    if not os.path.isdir(dirpath):
        _Dirs[dirpath] = 'not exist'
        if callback:
            reactor.callLater(0, callback, 'not exist', arg)  # @UndefinedVariable
        return 'not exist'
    d = threads.deferToThread(bpio.getDirectorySize, dirpath)
    d.addCallback(done, dirpath)
    _Jobs[dirpath] = (d, callback, arg)
    _Dirs[dirpath] = 'counting size'
    return 'counting size'


def done(size, dirpath):
    """
    This is called after you did ``ask`` of dir size.
    """
    global _Dirs
    global _Jobs
    if _Debug:
        lg.out(_DebugLevel, 'dirsize.done %s %s' % (str(size), dirpath.decode()))
    _Dirs[dirpath] = str(size)
    try:
        _, cb, arg = _Jobs.pop(dirpath, (None, None, None))
        if cb:
            cb(dirpath, size, arg)
    except:
        lg.exc()


def get(dirpath, default=''):
    """
    Only return directory size stored in memory - after previous calls to ``ask`` procedure.
    """
    global _Dirs
    return _Dirs.get(dirpath, default)


def isjob(dirpath):
    """
    You can check if some work is still in progress to calculate given folder
    size.
    """
    global _Jobs
    return dirpath in _Jobs


def getLabel(dirpath):
    """
    A very smart way to show folder size - can change units, must use ``ask`` first.
    """
    global _Dirs
    s = _Dirs.get(dirpath, '')
    if s not in ['counting size', 'not exist']:
        try:
            return diskspace.MakeStringFromBytes(int(s))
        except:
            return str(s)
    return str(s)


def getInBytes(dirpath, default=-1):
    """
    Return directory size in bytes, must use ``ask`` first.
    """
    return diskspace.GetBytesFromString(get(dirpath), default)


#------------------------------------------------------------------------------


def main():
    """
    Run the test - use command line to pass a location.
    """
    def _done(path, sz, *args, **kwargs):
        print(path, sz)
        reactor.stop()  # @UndefinedVariable

    bpio.init()
    ask(sys.argv[1], _done)
    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    main()
